/*	target/scsi_target.c

 	This file has all the functions for scsi midlevel on target side

*/
/*
	Copyright (C) 2001-2004 InterOperability Lab (IOL)
					University of New Hampshier (UNH)
					Durham, NH 03824

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
	USA.

	The name of IOL and/or UNH may not be used to endorse or promote products
	derived from this software without specific prior written permission.
*/

/* df revised begin */
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif

#ifndef KERNEL_VERSION			/* pre-2.1.90 didn't have it */
#define KERNEL_VERSION(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif
/* df revised end */

#include <linux/proc_fs.h>

#include "scsi_target.h"
#include "../common/lun_packing.h"

/*
# define DEBUG_REGISTER
# define DEBUG_DEREGISTER
# define DEBUG_RX_CMND
# define DEBUG_SCSI_DONE
# define DEBUG_ALLOCN_LEN
# define DEBUG_RX_DATA
# define DEBUG_HANDLE_CMD
# define DEBUG_SCSI_THREAD
*/

/* FUNCTION DECLARATIONS */

void scsi_target_process_thread(void *);

# if defined (FILEIO) || defined (GENERICIO)
static int build_filp_table(int);
static void close_filp_table(void);
# endif

# if defined (FILEIO) || defined (MEMORYIO)
static int get_inquiry_response(Scsi_Request *, __u32);
static int get_read_capacity_response(Scsi_Request *);
# endif

static int abort_notify(struct SM *);
static void aen_notify(int, __u64);

static int get_space(Scsi_Request *, __u32);
static int hand_to_front_end(Target_Scsi_Cmnd *);
static int handle_cmd(struct SC *);

# ifdef GENERICIO
static int fill_sg_hdr(Target_Scsi_Cmnd *, int);
void signal_process_thread(void *);
# endif

# ifdef DISKIO
static void te_cmnd_processed(Scsi_Cmnd *);
static int fill_scsi_device(void);
# endif

#ifdef CONFIG_PROC_FS
static int scsi_target_proc_info(char *buffer, char **start, off_t offset,
								 int length);
static int proc_scsi_target_gen_write(struct file *file, const char *buf,
									  unsigned long length, void *data);
static void build_proc_target_dir_entries(Scsi_Target_Device * the_device);
#endif

struct proc_dir_entry *proc_scsi_target;

/* 
 * Scsi_Target_Device: Maximum number of "active" targets at any time
 * Scsi_Target_Template: Maximum number of templates available
 *
 * The naming convention for variables is with the prefix "st" following
 * Eric's comments in the SCSI Initiator mid-level.
 * 
 * The idea with the following variables is to have a generic linked
 * list of templates - there is one template for one "type" of device
 * whereas for each device there is a single entry in the device list
 */

Target_Emulator target_data;

/*
 * okay the following is to get something off the ground and working.
 * In the mode that the emulator is talking to a real disk drive, there
 * will be a need to map the id given to us by rx_cmnd to host, channel,
 * id as far as the SCSI mid-level is concerned. For the present moment
 * we map all of these to the same target. So all commands are directed
 * to different LUNs on a single device. I visualize this as being a
 * linked list of scsi disks attached to the system and whenever a
 * command is received, the rx_cmnd function does some id_lookup and map
 * the individual commands to one in the linked list. How this mapping
 * gets done is an upper level management function that I do not want to
 * deal with yet	ASHISH
 */
int target_num;
Scsi_Device *scsi_device[MAX_TARGETS][MAX_LUNS];
int target_info[MAX_TARGETS][3];

# ifdef GENERICIO
char device[BYTE];
# endif

# ifdef FILEIO
char device[BYTE * 5];
# endif

/* 
 * scsi_target_init_module: Initializes the SCSI target module
 * FUNCTION: carry out all the initialization stuff that is needed
 * INPUT: none
 * OUTPUT: 0: everything is okay
 * 	      else trouble
 */
int
scsi_target_init_module(void)
{
	struct proc_dir_entry *generic;

	debugInit("scsi_target_init_module enter ... \n");
	/* initialize the global target_data struct */
	/*
	 * initialize the target semaphore as locked because you want
	 * the thread to block after entering the first loop
	 */
	init_MUTEX_LOCKED(&target_data.target_sem);
	init_MUTEX(&target_data.queue_sem);
	init_MUTEX_LOCKED(&target_data.thread_sem);

	target_data.add_delete = SPIN_LOCK_UNLOCKED;
	target_data.msg_lock = SPIN_LOCK_UNLOCKED;
	target_data.st_device_list = NULL;
	target_data.st_target_template = NULL;
	target_data.cmd_queue_start = NULL;
	target_data.cmd_queue_end = NULL;
	target_data.msgq_start = NULL;
	target_data.msgq_end = NULL;
	target_data.command_id = 0;
	target_data.thread_id = NULL;

# ifdef MEMORYIO
	target_num = 1;
# endif

# ifdef DISKIO
	if (fill_scsi_device() == -1) {
		printk("scsi_target_init_module: fill_scsi_device returned an error\n");
		return -1;
	}
# endif

	/*
	 * create a list of devices that are available. This is a very
	 * simplistic mapping of these devices. Each new device will
	 * represent a new LUN. A new mapping function will have to
	 * solve this problem - LATER - Ashish
	 */
# ifdef GENERICIO
	if (build_filp_table(IOGENERIC) < 0) {
		printk("sc..init: build_filp_table returned an error\n");
		return -1;
	}
# endif

# ifdef FILEIO
	if (build_filp_table(IOFILE) < 0) {
		printk("sc..init: build_filp_table returned an error\n");
		return -1;
	}
# endif

#ifdef CONFIG_PROC_FS
	proc_scsi_target = proc_mkdir("scsi_target", 0);
	if (!proc_scsi_target) {
		printk(KERN_ERR "cannot init /proc/scsi_target\n");
		return -ENOMEM;
	}
	generic =
		create_proc_info_entry("scsi_target/scsi_target", 0, 0,
							   scsi_target_proc_info);
	if (!generic) {
		printk(KERN_ERR "cannot init /proc/scsi_target/scsi_target\n");
		remove_proc_entry("scsi_target", 0);
		return -ENOMEM;
	}
	generic->write_proc = proc_scsi_target_gen_write;
#endif

	/* spawn a thread */
	kernel_thread((int (*)(void *)) scsi_target_process_thread, NULL, 0);
	debugInit("exit\n");
	return 0;
}

/* 
 * scsi_target_cleanup_module: Removal of the module from the kernel
 * FUNCTION: carry out the cleanup stuff
 * INPUT: None allowed
 * OUTPUT: None allowed
 */
void
scsi_target_cleanup_module(void)
{
# ifdef GENERICIO

	/* close all the devices that you have opened */
	close_filp_table();

	if (target_data.signal_id != NULL) {
		send_sig(SIGKILL, target_data.signal_id, 1);
		down_interruptible(&target_data.signal_sem);
	}
# endif

# ifdef FILEIO
	close_filp_table();
# endif

#ifdef CONFIG_PROC_FS
	/* Don't show the /proc/scsi_target files */
	remove_proc_entry("scsi_target/scsi_target", 0);
	remove_proc_entry("scsi_target", 0);
#endif

	/* kill the thread */
	if (target_data.thread_id != NULL) {
		send_sig(SIGKILL, target_data.thread_id, 1);
		down_interruptible(&target_data.thread_sem);
	}
	printk("Exiting scsi module\n");
}

module_init(scsi_target_init_module);
module_exit(scsi_target_cleanup_module);

/*
 * register_target_template: to register a template with the midlevel
 * FUNCTION: to add the midlevel template to the linked list
 * INPUT: pointer to the template
 * OUTPUT: 0 for successful registration
 * 	   < 0 for trouble
 */
int
register_target_template(Scsi_Target_Template * the_template)
{
	int check;
	Scsi_Target_Template *st_current;

	if (!the_template) {		/* huh */
		printk
			("register_target_template: Refusal to register a NULL template\n");
		return -1;
	}

	the_template->device_usage = 0;

	/* check if the template is already in the list */
	for (st_current = target_data.st_target_template; st_current != NULL;
		 st_current = st_current->next) {
		/* loop */
# ifdef DEBUG_REGISTER
		printk("Looping template\n");
# endif
		if (!strcmp(target_data.st_target_template->name, the_template->name))
			break;
	}

	if (st_current != NULL) {	/* template is old */
		printk("register_target_template: Template already present\n");
		return -1;
	}

	the_template->next = target_data.st_target_template;
	target_data.st_target_template = st_current = the_template;

# ifdef DEBUG_REGISTER
	printk("module count increased\n");
# endif

/* Ming Zhang, mingz@ele.uri.edu */
#ifndef K26
	MOD_INC_USE_COUNT;
#endif

	/* check if the device has a detect function */
	if (st_current->detect) {
		/* detect the device */
		check = st_current->detect(the_template);
		if (check < 0) {		/* error */
			printk("Hosts detected: %d .. removing template\n", check);
			deregister_target_template(the_template);
			return check;
		} else {
			printk("Hosts detected: %d ... incrementing device usage\n", check);
		}
	} else {
		printk
			("register_target_template: template does not have a detect function\n");
		deregister_target_template(the_template);
		return -1;
	}

	return 0;
}

/*
 * deregister_target_template:
 * FUNCTION: removal of the target template from the list
 * INPUT: pointer to STT
 * OUTPUT: 0 for successful deregistration
 * 	   else < 0 for trouble
 */
int
deregister_target_template(Scsi_Target_Template * the_template)
{
	Scsi_Target_Template *st_current, *st_prev = NULL;
	Scsi_Target_Device *st_dev_curr;

	st_dev_curr = target_data.st_device_list;

	if (!the_template) {		/* huh */
		printk("deregister_target_template: Cannot remove a NULL template\n");
		return -1;
	}

	/* remove devices that are using this template */
	while (st_dev_curr && st_dev_curr->template
		   && (st_dev_curr->template->device_usage > 0)) {
		st_dev_curr = target_data.st_device_list;
		while (st_dev_curr) {
			if (!strcmp(st_dev_curr->template->name, the_template->name)) {
# ifdef DEBUG_DEREGISTER
				printk("dereg...tmpt: Match found for %s%d\n",
					   st_dev_curr->template->name, (int) st_dev_curr->id);
# endif
				deregister_target_front_end(st_dev_curr);
				st_dev_curr = target_data.st_device_list;	// Edward
				break;
			} else {
				st_dev_curr = st_dev_curr->next;
				if (st_dev_curr) {
					printk
						("dereg..tmpt: Error ... no device found with the required template %s\n",
						 the_template->name);
					return -1;
				}
			}
		}
	}

	/* remove template from device list */
	for (st_current = target_data.st_target_template; st_current != NULL;
		 st_current = st_current->next) {
# ifdef DEBUG_DEREGISTER
		printk("Looping dereg template\n");
# endif
		if (!strcmp(st_current->name, the_template->name)) {
# ifdef DEBUG_DEREGISTER
			/* match found */
			printk("deregister_template: template match found\n");
# endif
			/* check if there are any devices using this template */
			if (!st_current->device_usage) {
				/* this template can be removed */
				if (!st_prev)	/* First element */
					target_data.st_target_template =
						target_data.st_target_template->next;
				else			/* middle of the road */
					st_prev->next = st_current->next;
				break;
			} else {			/* A device still uses the template */
				printk("scsi_target: Non-zero device usage ouch !!\n");
				return -1;	/* should never get here */
			}
		} else
			st_prev = st_current;
	}

# ifdef DEBUG_DEREGISTER
	printk("Decreasing module count\n");
# endif

/* Ming Zhang, mingz@ele.uri.edu */
#ifndef K26
	MOD_DEC_USE_COUNT;
#endif

	return 0;
}

/*
 * register_target_front_end:
 * FUNCTION: 	to register the individual device with the mid-level
 * 		to allocate the device an device id etc
 * 		start a thread that will be responsible for the target
 * INPUT:	pointer to a struct STD
 * OUTPUT:	Scsi_Target_Device - if everything is okay
 * 		else NULL if there is trouble
 */
Scsi_Target_Device *
register_target_front_end(Scsi_Target_Template * tmpt)
{
	Scsi_Target_Device *the_device;

	the_device =
		(Scsi_Target_Device *) kmalloc(sizeof (Scsi_Target_Device), GFP_KERNEL);
	if (!the_device) {
		printk
			("register_target_front_end: Could not allocate space for the device\n");
		return NULL;
	}

	if (!tmpt) {
		printk
			("register_target_front_end: Cannot register NULL device template !!!\n");
		return NULL;
	}

	/* fill up the struct */
	the_device->template = tmpt;

	the_device->next = target_data.st_device_list;

	if (the_device->next)
		the_device->id = the_device->next->id + 1;
	else
		the_device->id = 0;		/* first device */

	target_data.st_device_list = the_device;

# ifdef DEBUG_DEREGISTER
	printk("reg..end: device %s%d added\n", tmpt->name, (int) the_device->id);
# endif

	/* Added by naren for proc support */
#ifdef CONFIG_PROC_FS
	build_proc_target_dir_entries(the_device);
#endif

	tmpt->device_usage++;
	return the_device;
}

/*
 * deregister_target_front_end:
 * FUNCTION:	to allow removal of the individual device from the
 * 		midlevel
 * 		free up the device id number for future use
 * 		close the thread responsible for the target after making
 * 		sure that all existing commands have been responded to
 * 		we need to do serious error checking since this can be
 * 		called by different front-ends
 * 		CANNOT BE CALLED FROM INTERRUPT CONTEXT
 * INPUT:	pointer to a struct STD to be removed
 * OUTPUT:	int - 0 if everything is okay
 * 		else < 0 if there is trouble
 */
int
deregister_target_front_end(Scsi_Target_Device * the_device)
{
	Scsi_Target_Device *curr, *previous = NULL;
	Target_Scsi_Cmnd *cmnd;

	if (!the_device) {
		printk
			("dereg...end: cannot remove NULL devices corresponding to a NULL template\n");
		return -1;
	}

	if (!the_device->template->device_usage) {
		printk
			("dereg...end: 0 device usage and a device to deregister ... a contradiction me thinks\n");
	}

	/* 
	 * go through the device list till we get to this device and
	 * then remove it from the list
	 */
	for (curr = target_data.st_device_list; curr != NULL; curr = curr->next) {
		if (curr == the_device) {
# ifdef DEBUG_DEREGISTER
			printk("dereg..end: We have a match\n");
# endif
			break;
		} else
			previous = curr;
	}

	if (!curr) {				/* No match found */
		printk("dereg..end: No match found\n");
		return -1;
	}

	/* remove it from the list */
	if (previous)				/* not the first device */
		previous->next = curr->next;

	else
		target_data.st_device_list = curr->next;

	/* release the device */
	if (curr->template->release) {
		if (curr->template->release(curr)) {
			printk("dereg...end: release of device failed\n");
			return -1;
		}
	}

	/* drop all commands corresponding to this device */
	if (!down_interruptible(&target_data.queue_sem)) {
		for (cmnd = target_data.cmd_queue_start; cmnd != NULL; cmnd = cmnd->next) {
			if (cmnd->dev_id == curr->id)
				cmnd->state = ST_DEQUEUE;
		}
		up(&target_data.queue_sem);
	}

	/* wake up the mid-level thread so it can dequeue stuff */
	up(&target_data.target_sem);

	/* reduce device usage */
	curr->template->device_usage--;

	/* freeing things */
	kfree(curr);
	curr = NULL;

	return 0;
}

# ifdef GENERICIO
# define SETFL_MASK (O_APPEND | O_NONBLOCK | O_NDELAY | FASYNC)

/*
 * signal_process_thread: This thread is responsible for receiving SIGIO
 * signals when SCSI generic is used for processing. This receives the
 * SIGIO signal, changes the state of the oldest command that is in the
 * ST_PROCESSING state and goes back to sleep. I found this to be
 * necessary because signals tend to screw up the processing of the scsi
 * generic commands. Hopefully this will clear things up
 */
void
signal_process_thread(void *param)
{
	devset_t *temp;
	int i, sigioflag = 0;
	struct file *dev_file;
	Target_Scsi_Cmnd *cmd_curr;
	unsigned long arg;

	lock_kernel();

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();
	sprintf(current->comm,
#endif
		"signal_thread");

	/* also we want to be able to receive SIGIO */
	siginitsetinv(&current->blocked, IO_SIGS);

	target_data.signal_id = current;

	unlock_kernel();

	/* 
	 * we must now prepare this thread to receive the SIGIO signal
	 * for the opened SCSI device(s)
	 */
	for (temp = target_data.fdlist; temp != NULL; temp = temp->next) {
		printk("1 ...");
		dev_file = temp->device;
		if (dev_file) {
			arg = dev_file->f_flags | FASYNC;
			lock_kernel();

			/* equal to fcntl (fd, F_SETOWN, getpid()) */
			dev_file->f_owner.pid = current->pid;
			dev_file->f_owner.uid = current->uid;
			dev_file->f_owner.euid = current->euid;

			/* equal to fcntl (fd,F_SETFL,flags|O_ASYNC) */
			if ((arg ^ dev_file->f_flags) & FASYNC) {
				if (dev_file->f_op && dev_file->f_op->fasync) {
					i = dev_file->f_op->fasync(0, dev_file,
											   (arg & FASYNC) != 0);
					if (i < 0) {
						printk("si..thread: fasync returned %d\n", i);
						return;
					}
				}
			}

			/* required for strict SunOS emulation */
			/*
			   if (O_NONBLOCK != O_NDELAY)
			   if (arg & O_NDELAY)
			   arg |= O_NONBLOCK;
			 */
			dev_file->f_flags =
				(arg & SETFL_MASK) | (dev_file->f_flags & ~SETFL_MASK);
			unlock_kernel();
		} else {
			printk
				("signal_process_thread: Could not file a device to write to\n");
			return;
		}
	}

	while (1) {
		down_interruptible(&target_data.sig_thr_sem);

		sigioflag = 0;
		if (signal_pending(current)) {
			spin_lock_irq(&current->sigmask_lock);
			if (sigismember(&current->pending.signal, SIGIO)) {
# ifdef DEBUG_SCSI_THREAD
				printk("SIGIO received\n");
# endif
				sigioflag = 1;
				sigdelset(&current->pending.signal, SIGIO);
			}
			recalc_sigpending(current);
			spin_unlock_irq(&current->sigmask_lock);
			if (!sigioflag) {
				printk("si...thread: time to die\n");
				break;
			}
		}

		cmd_curr = target_data.cmd_queue_start;

		while (cmd_curr) {
			if (cmd_curr->state == ST_PROCESSING) {
				cmd_curr->state = ST_PROCESSED;
				up(&target_data.target_sem);
				/* 
				 * change the state of all PROCESSING
				 * commands ???
				 */
//              break;
			}
			cmd_curr = cmd_curr->next;
		}
	}
	up(&target_data.signal_sem);
}
# endif

/*
 * scsi_target_process_thread: this is the mid-level target thread that
 * is responsible for processing commands.
 */
void
scsi_target_process_thread(void *param)
{
	int i;
# ifdef GENERICIO
	mm_segment_t old_fs;
	struct file *dev_file;
# endif
	struct scatterlist *st_list;
	Target_Scsi_Cmnd *cmd_curr, *cmd_copy, *cmd_prev = NULL;
	Target_Scsi_Message *msg;
	unsigned long flags;
	int lun;
# ifdef DISKIO
	Scsi_Device *this_device = NULL;
# endif

	lock_kernel();

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();
	sprintf(current->comm,
#endif
		"target_thread");

	siginitsetinv(&current->blocked, SHUTDOWN_SIGS);

	target_data.thread_id = current;

	unlock_kernel();

# ifdef DEBUG_SCSI_THREAD
	printk("STARTING SCSI_TARGET_THREAD\n");
# endif

	while (1) {
		if (down_interruptible(&target_data.target_sem))
			goto scsi_thread_out;

# ifdef DEBUG_SCSI_THREAD
		printk("scsi_target_process_thread awake\n");
# endif

		/* is message received */
		while (target_data.msgq_start) {
			/* house keeping */
			spin_lock_irqsave(&target_data.msg_lock, flags);
			msg = target_data.msgq_start;
			target_data.msgq_start = msg->next;
			if (msg->next)
				msg->next->prev = NULL;
			if (target_data.msgq_end == msg)
				target_data.msgq_end = NULL;
			spin_unlock_irqrestore(&target_data.msg_lock, flags);

			/* execute function */
			switch (msg->message) {
			case ABORT_TASK:
				{
					Target_Scsi_Cmnd *cmnd;
					cmnd = (Target_Scsi_Cmnd *) msg->value;
					for (cmd_curr = target_data.cmd_queue_start;
						 cmd_curr != NULL; cmd_curr = cmd_curr->next) {
						if ((cmd_curr->id == cmnd->id)
							&& (cmd_curr->lun == cmnd->lun))
							break;
					}

					if (cmd_curr) {
						cmd_curr->abort_code = CMND_ABORTED;
						//if (cmd_curr->state != ST_PROCESSING)
						//	cmd_curr->state = ST_DEQUEUE;
						if (abort_notify(msg)) {
							printk("sc..thread: err for abort command %d\n",
								   cmd_curr->id);
							goto scsi_thread_out;
						}
					} else
						printk
							("scsi..thread: Could not find command in list\n");
					break;
				}

			case LUN_RESET:
				{
					/* BAD BAD REALLY BAD */
					__u64 lun = *((__u64 *) msg->value);
					for (cmd_curr = target_data.cmd_queue_start;
						 cmd_curr != NULL; cmd_curr = cmd_curr->next) {
						if (cmd_curr->lun == lun)
							scsi_release(cmd_curr);
					}

					aen_notify(msg->message, lun);
					break;
				}

			case TARGET_RESET:
				{
					for (cmd_curr = target_data.cmd_queue_start;
						 cmd_curr != NULL; cmd_curr = cmd_curr->next)
						scsi_release(cmd_curr);
					break;

					aen_notify(msg->message, 0);
				}

			default:
				{
					printk("scsi..thread: Should never get here\n");
					break;
				}
			}

			kfree(msg);
			msg = NULL;
		}

# ifdef DEBUG_SCSI_THREAD
		printk("sc..thread: Searching command queue\n");
# endif
		cmd_curr = target_data.cmd_queue_start;
		cmd_prev = NULL;
		while (cmd_curr) {
# ifdef DEBUG_SCSI_THREAD
			printk("sc..thread: command %p id %d status %d\n", cmd_curr,
				   cmd_curr->id, cmd_curr->state);
# endif
			/* is command received */
			if (cmd_curr->state == ST_NEW_CMND) {
# ifdef DEBUG_SCSI_THREAD
				printk("sc..thread: New command %p id: %d\n", cmd_curr,
					   cmd_curr->id);
# endif

				lun = cmd_curr->lun;

# ifdef DISKIO
				if (lun < MAX_LUNS)
					this_device = scsi_device[cmd_curr->target_id][lun];

				if (!this_device) {
					printk
						("the lun %d is invalid, assign lun 0 to handle rsp\n",
						 lun);
					this_device = scsi_device[cmd_curr->target_id][0];
				}

				if (this_device) {
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
					cmd_curr->req = scsi_allocate_request(this_device, GFP_ATOMIC);
#else
					cmd_curr->req = scsi_allocate_request(this_device);
#endif
				} else {
					printk("sc..thread: Can't find the device\n");
					goto scsi_thread_out;
				}
# else
				cmd_curr->req =
					(Scsi_Request *) kmalloc(sizeof (Scsi_Request),
											 GFP_KERNEL | GFP_ATOMIC);

				if (cmd_curr->req)
					memset(cmd_curr->req, 0, sizeof (Scsi_Request));

				if (lun >= MAX_LUNS)
					cmd_curr->req->sr_allowed = 1;
# endif

				if (!cmd_curr->req) {
					printk("sc..thread: req: kmalloc failed\n");
					goto scsi_thread_out;
				}

				memcpy(cmd_curr->req->sr_cmnd, cmd_curr->cmd, cmd_curr->len);

				if (handle_cmd(cmd_curr)) {
					printk("sc..thread: handle_cmd: err for command %d\n",
						   cmd_curr->id);
					/* is bailing out a good idea */
					goto scsi_thread_out;
				}
			}

			/* is a command pending */
			if (cmd_curr->state == ST_PENDING) {
# ifdef DEBUG_SCSI_THREAD
				printk("sc..thread: Command %p id: %d pending\n", cmd_curr,
					   cmd_curr->id);
# endif
				/* call the rdy_to_xfer function */
				if (hand_to_front_end(cmd_curr)) {
					printk
						("sc..thread: hand_to_front_end: returned an error for command %d\n",
						 cmd_curr->id);
					goto scsi_thread_out;
				}
			}

			/* is data received */
			if (cmd_curr->state == ST_TO_PROCESS) {
				/*
				 * we have received the data - does this
				 * go off to handle_cmd again ?
				 */
# ifdef DEBUG_SCSI_THREAD
				printk("sc..thread: Command %p id: %d with data received\n",
					   cmd_curr, cmd_curr->id);
# endif
				if (handle_cmd(cmd_curr)) {
					printk("sc..thread: handle_cmd: err for command %d\n",
						   cmd_curr->id);
					/* is bailing out a good idea */
					goto scsi_thread_out;
				}
			}
# ifdef GENERICIO
			/* is command processed */
			if (cmd_curr->state == ST_PROCESSED) {
# ifdef DEBUG_SCSI_THREAD
				printk("sc..thread: Command %p id: %d processed\n", cmd_curr,
					   cmd_curr->id);
# endif
				/* we have some reading to do */
				dev_file = cmd_curr->fd->device;
				if ((dev_file) && (dev_file->f_op) && (dev_file->f_op->read)) {
					old_fs = get_fs();
					set_fs(get_ds());
					i = dev_file->f_op->read(dev_file,
											 (__u8 *) cmd_curr->sg,
											 sizeof (sg_io_hdr_t),
											 &dev_file->f_pos);
					set_fs(old_fs);
					if ((i < 0) && (i != -EAGAIN)) {
						printk("sc..thread: read returned an error %d\n", i);
						goto scsi_thread_out;
					}
					cmd_curr->state = ST_DONE;
				}
			}
# endif
			/* is command done */
			if (cmd_curr->state == ST_DONE) {
				/* 
				 * hand it off to the front end driver
				 * to transmit
				 */

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
#ifdef DISKIO
				/*
				 * for DISKIO under K26, should reassign sg_dma_address here
				 */
				for (i = 0; i < cmd_curr->req->sr_use_sg; i++) {
					sg_dma_address(((struct scatterlist *)cmd_curr->req->sr_buffer) + i) = 
						(dma_addr_t)page_address(((struct scatterlist *)cmd_curr->req->sr_buffer)[i].page);
					((struct scatterlist *)cmd_curr->req->sr_buffer)[i].offset = (unsigned long)
						sg_dma_address(((struct scatterlist *)cmd_curr->req->sr_buffer) + i) & ~PAGE_MASK;
				}
#endif
#endif

# ifdef DEBUG_SCSI_THREAD
				printk("sc..thread: Command %p id: %d done\n", cmd_curr,
					   cmd_curr->id);
# endif
				if (hand_to_front_end(cmd_curr)) {
					printk
						("sc..thread: hand_to_front_end: returned an error for command %d\n",
						 cmd_curr->id);
					goto scsi_thread_out;
				}
			}

			/* can command be dequeued */
			if (cmd_curr->state == ST_DEQUEUE) {
				/*
				 * dequeue the command and free it
				 */
# ifdef DEBUG_SCSI_THREAD
				printk("sc..thread: Command %p id: %d - to dequeue\n", cmd_curr,
					   cmd_curr->id);
				if (cmd_curr->next)
					printk("sc..thread: Command %p id: %d - in queue\n",
						   cmd_curr->next, cmd_curr->next->id);
# endif
				/* so that we can repoint cmd_curr */
				cmd_copy = cmd_curr;

				/* dequeue the command */
				spin_lock_irqsave(&target_data.add_delete, flags);
				if (cmd_prev) {
					cmd_prev->next = cmd_curr->next;
					if (cmd_curr == target_data.cmd_queue_end)	/* Last element */
						target_data.cmd_queue_end = cmd_prev;
				} else {			/* first element */
					if (target_data.cmd_queue_end == cmd_curr) {	/* only element */
						target_data.cmd_queue_end = NULL;
						target_data.cmd_queue_start = NULL;
					} else
						target_data.cmd_queue_start = cmd_curr->next;
				}

				/* reset pointers */
				cmd_curr = cmd_curr->next;
				spin_unlock_irqrestore(&target_data.add_delete, flags);
# ifdef GENERICIO
				/* free up the SCSI generic stuff */
				if (cmd_copy->sg->dxferp)
					kfree(cmd_copy->sg->dxferp);
				kfree(cmd_copy->sg->sbp);
				kfree(cmd_copy->sg);
# endif
				/* free up pages */
				st_list = (struct scatterlist *) cmd_copy->req->sr_buffer;
				for (i = 0; i < cmd_copy->req->sr_use_sg; i++) {
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
					kunmap(st_list[i].page);
					__free_page(st_list[i].page);
#else
					free_page((long int) st_list[i].address);
#endif
				}
				
				/* free up scatterlist */
				if (cmd_copy->req->sr_use_sg)
					kfree(st_list);

				/* free up Scsi_Request */
# ifdef DISKIO
				scsi_release_request(cmd_copy->req);
# else
				kfree(cmd_copy->req);
# endif
				/* free up Target_Scsi_Cmnd */
				kfree(cmd_copy);

				cmd_copy = NULL;
			}

			/* if we get through to here then move on */
			else if (cmd_curr) {
				/* proceed to the next command */
# ifdef DEBUG_SCSI_THREAD
				printk("sc..thread: Moving to the command after %p, id: %d\n",
					   cmd_curr, cmd_curr->id);
# endif
				cmd_prev = cmd_curr;
				cmd_curr = cmd_curr->next;
			}
		}
# ifdef DEBUG_SCSI_THREAD
		printk("sc..thread: going back to sleep again\n");
# endif
	}
  scsi_thread_out:
	printk("%s exiting\n", current->comm);
	up(&target_data.thread_sem);
	return;
}

/*
 * rx_cmnd: this function is the basic function called by any front end
 * when it receives a SCSI command. The rx_cmnd function then fills up
 * a Target_Scsi_Cmnd struct, adds to the queue list, awakens the mid-
 * level thread and then returns the struct to the front end driver
 * INPUT:	device, target_id, lun, SCSI CDB as an unsigned char
 * 		length of the command (or size of scsi_cdb array if 
 * 		unavailable
 * OUTPUT:	Target_Scsi_Cmnd struct or NULL if things go wrong
 */
Target_Scsi_Cmnd *
rx_cmnd(Scsi_Target_Device * device, __u64 target_id,
	__u64 lun, __u8 *scsi_cdb, int len, int datalen, int in_flags,
	Target_Scsi_Cmnd **result_command)
{
	Target_Scsi_Cmnd *command;
	unsigned long flags;
# if defined (GENERICIO) || defined (FILEIO)
	devset_t *temp;
# endif

	*result_command = NULL;
	if (!target_data.thread_id) {
		printk("rx_cmnd: No Mid-level running !!!!\n");
		return NULL;
	}

	if (!device) {
		printk("rx_cmnd: No device given !!!!\n");
		return NULL;
	}

	*result_command = command =
		(Target_Scsi_Cmnd *) kmalloc(sizeof (Target_Scsi_Cmnd),
						GFP_KERNEL | GFP_ATOMIC);
	if (command == NULL) {
		printk("rx_cmnd: kmalloc: command ... No space\n");
		/* sendsig (SIGKILL, target_data.thread_id, 0); */
		return NULL;
	}
# ifdef DEBUG_RX_CMND
	printk("rx_cmnd: filling up command struct %p\n", command);
# endif

	/* fill in Target_Scsi_Cmnd */
	memcpy(command->cmd, scsi_cdb, MAX_COMMAND_SIZE);
	command->state = ST_NEW_CMND;
	command->abort_code = CMND_OPEN;
	command->device = device;
	command->dev_id = device->id;

	/*ramesh@global.com
	   added data length and flgs to the command structure */
	command->datalen = datalen;
	command->flags = in_flags;

	/* cdeng change target_id later if lun doesn't match */
	command->target_id = target_id;
	command->lun = unpack_lun((__u8 *) &lun);
	command->next = NULL;
	if ((len < MAX_COMMAND_SIZE) && (len > 0))
		command->len = len;
	else {
//          printk ("setting cmd len to %d instead of %d\n", MAX_COMMAND_SIZE, len);
		command->len = MAX_COMMAND_SIZE;
	}

# if defined (FILEIO) || defined (GENERICIO)
	/* fill in the device descriptor */
	command->fd = NULL;
	for (temp = target_data.fdlist; temp != NULL; temp = temp->next) {
# ifdef DEBUG_RX_CMND
		printk
			("rx_cmnd: looking for id: %d, lun: %d: in loop for id: %d, lun: %d\n",
			 (int) command->target_id, (int) command->lun,
			 (int) temp->target_id, (int) temp->lun);
# endif

		if ((command->target_id == temp->target_id) &&
			(command->lun == temp->lun)) {
# ifdef DEBUG_RX_CMND
			printk("rx_cmnd: Match found\n");
# endif
			command->fd = temp;
			break;
		}
	}

	if (command->fd == NULL) {
		command->fd = target_data.fdlist;
		printk
			("rx_cmnd: Could not find a match for the given command with id: %d, lun:%d\n",
			 (int) command->target_id, (int) command->lun);
		printk
			("rx_cmnd: Let last lun handle it now and change to not-supported lun later\n");
		//kfree (command);
		//return NULL;
	}
# endif

	/*
	 * this could potentially be gross
	 */
	spin_lock_irqsave(&target_data.add_delete, flags);
# ifdef DEBUG_RX_CMND
	printk("rx_cmnd: spinlock for %p\n", command);
# endif
	command->id = ++target_data.command_id;
	/* check this to make sure you dont have a command with this id ?????
	 * IGNORE FOR NOW
	 */
	if (!command->id) {
		/* FOR WRAP AROUNDS */
		command->id = ++target_data.command_id;
	}

	if ((!target_data.cmd_queue_start) && (!target_data.cmd_queue_end)) {
		target_data.cmd_queue_start = target_data.cmd_queue_end = command;
	} else {
		target_data.cmd_queue_end->next = command;
		target_data.cmd_queue_end = command;
	}

# ifdef DEBUG_RX_CMND
	printk("rx_cmnd: spinlock for %p with id %d\n", command, command->id);
# endif

	spin_unlock_irqrestore(&target_data.add_delete, flags);

	/* wake up thread */
	up(&target_data.target_sem);
	return command;
}

/*
 * scsi_rx_data: This function is called by the lower-level driver to 
 * tell the mid-level that data corresponding to a command has been 
 * received. This function can be called from within an interrupt 
 * handler (??). All this function does currently is to change the 
 * state of a command and then wake the mid-level thread to deal with 
 * this command.
 * INPUT: scsi command for which data has been received
 * OUTPUT: 0 if okay else < 0
 */
int
scsi_rx_data(Target_Scsi_Cmnd * the_command)
{
	Target_Scsi_Cmnd *curr = target_data.cmd_queue_start;

	if (!the_command) {
		printk("rx_data: cannot deal with NULL command\n");
		return -1;
	}

	/* check to see if the command exists in the queue */
	while (curr) {
		if (curr->id == the_command->id) {
# ifdef DEBUG_RX_DATA
			printk("scsi_rx_data: Match found\n");
# endif
			curr->state = ST_TO_PROCESS;
			break;
		}
		curr = curr->next;
	}

	if (!curr) {				/* this command may have been aborted */
		printk("rx_data: could not find a match for the command\n");
		printk
			("rx_data: This may not be serious ... the command may have been aborted\n");
		return -1;
	}

	/* wake up the mid-level thread */
	up(&target_data.target_sem);

	return 0;
}

/*
 * scsi_target_done: This is the function called by the low-level driver
 * to signify that it has completed the execution of a given scsi cmnd
 * This function needs to remove the resources that have been allocated
 * to the given command, dequeue the command etc. This function will be
 * called from within the context of an interrupt handler. So, it may
 * be best to leave it up to the mid-level thread to actually deal with
 * these functions. Right now, I am setting it up so that the status of
 * the command is changed and the mid-level is awakened.
 * INPUT: scsi command to be dequeued
 * OUTPUT: 0 if everything is okay
 * 	   < 0 if there is trouble
 */
int
scsi_target_done(Target_Scsi_Cmnd * the_command)
{
	Target_Scsi_Cmnd *curr = target_data.cmd_queue_start;

	if (!the_command) {
		printk("scsi_target_done: cannot dequeue NULL command\n");
		return -1;
	}

	/* check to see if the command exists in the queue */
	while (curr) {
		if (curr->id == the_command->id) {
# ifdef DEBUG_SCSI_DONE
			printk("scsi_target_done: Match found\n");
# endif
			curr->state = ST_DEQUEUE;
			break;
		}
		curr = curr->next;
	}

	if (!curr) {				/* should never happen */
		printk("scsi_target_done: could not find a match for the command\n");
		/*
		 * This error may not be so critical. We should be able
		 * to do away with this command without affecting things
		 * For now kill everything
		 */
		return -1;
	}

	/* in the clear - awaken the scsi_thread to do the needful */
	up(&target_data.target_sem);

	return 0;
}

/*
 * scsi_release: This function is called by a low-level driver when it
 * determines that it does not responses to certain commands. Situations
 * like this happen when for instance a LIP is received in a Fibre
 * Channel Loop or when a Logout is received in iSCSI before command
 * execution is completed. The low-level driver may no longer care about
 * receiving responses for those commands. This function can be called
 * from within interrupt context
 * INPUT: command to release
 * OUTPUT: 0 if success, < 0 if there is trouble
 */
int
scsi_release(Target_Scsi_Cmnd * cmnd)
{
	Target_Scsi_Cmnd *curr = target_data.cmd_queue_start;

	if (!cmnd) {
		printk("scsi_release: Cannot release a NULL command!!!\n");
		return -1;
	}

	/* check to see if the command exists in our queue */
	while (curr) {
		if (curr->id == cmnd->id) {
			curr->abort_code = CMND_RELEASED;
			/* 
			 * if a command is processing, it is not nice to
			 * dequeue a command, as we will get a response
			 * anyways. This may be an inherent race condn
			 * We catch what we can and move on. A second 
			 * check performed when hand_to_front_end is
			 * called should catch the remaining commands
			 * - Ashish
			 */
			if (curr->state != ST_PROCESSING)
				curr->state = ST_DEQUEUE;
			break;
		}
		curr = curr->next;
	}

	if (!curr) {
		printk("scsi_release: No match for command with id %d\n",
			cmnd->id);
		return -1;
	}

	up(&target_data.target_sem);
	return 0;
}

/*
 * rx_task_mgmt_fn: This function is called by a low-level driver to
 * indicate to the Mid-level that it has received a management function.
 * This function will decide the action to be taken in response to this
 * management function. This function will in turn create a message list
 * in the mid-level. CAN BE CALLED FROM INTERRUPT CONTEXT
 * INPUT: device, function, command - if relevant
 * OUTPUT: message or NULL if there is trouble
 * Definition of value:
 * 1. ABORT
 * 	value = Target_Scsi_Cmnd
 * 2. ABORT TASK SET
 * 	This function cannot be directly performed since the Mid-Level
 * 	has no knowledge of what Initiators are logged in to the front-
 * 	end. This functionality can be achieved by issuing ABORTS to 
 * 	various commands identified within the Task Set by the front
 * 	-end. value = N/A
 * 3. CLEAR ACA
 * 	I dont understand this one
 * 4. CLEAR TASK SET
 * 	Implement similar to ABORT Task Set - I know that all semantics 
 * 	for this cannot be implemented (Any ideas ?)
 * 5. LUN RESET
 * 	set value = pointer to LUN
 * 6. TARGET RESET
 * 	set value = NULL
 */
struct SM *
rx_task_mgmt_fn(struct STD *dev, int fn, void *value)
{
	unsigned long flags;
	Target_Scsi_Message *msg;

	msg =
		(Target_Scsi_Message *) kmalloc(sizeof (Target_Scsi_Message),
										GFP_KERNEL | GFP_ATOMIC);
	if (!msg) {
		printk
			("rx_task_mgmt_fn: could not allocate space for the scsi message\n");
		return NULL;
	}

	if ((fn < ABORT_TASK) && (fn > TARGET_RESET)) {
		printk("rx_task_mgmt_fn: Invalid value for Task Management function\n");
		return NULL;
	}

	if ((fn == ABORT_TASK_SET) || (fn == CLEAR_ACA) || (fn == CLEAR_TASK_SET)) {
		printk
			("rx_task_mgmt_fn: This task management function is not implemented. See the comments above the rx_task_mgmt_fn definition\n");
		return NULL;
	} else if ((fn == ABORT) && (value == NULL)) {
		printk("rx_task_mgmt_fn: Cannot abort a NULL command\n");
		return NULL;
	} else if ((fn == LUN_RESET) || (fn == TARGET_RESET)) {
		if (value != NULL) {	/* not a critical error */
			printk
				("rx_task_mgmt_fn: setting a value here is irrelevant. Ignoring this value and executing the function anyways\n");
		}
	}

	msg->next = NULL;
	msg->prev = NULL;
	msg->device = dev;
	msg->value = value;
	msg->message = fn;

	spin_lock_irqsave(&target_data.msg_lock, flags);

	if (!target_data.msgq_start) {
		target_data.msgq_start = target_data.msgq_end = msg;
	} else {
		target_data.msgq_end->next = msg;
		target_data.msgq_end = msg;
	}

	spin_unlock_irqrestore(&target_data.msg_lock, flags);

	/* wake up thread */
	up(&target_data.target_sem);
	return msg;
}

# if defined (FILEIO) || defined (GENERICIO)
/*
 * build_filp_table: builds up a table of open file descriptors.
 * INPUT: int (IOGENERIC/IOFILE)
 * OUTPUT: no of devices opened or < 0 if there is trouble
 */
static int
build_filp_table(int type)
{
	char *tmp;
	int error, i;
	mm_segment_t old_fs;
	devset_t *temp;
	struct file *dev_file;

	target_num = 1;
	target_data.fdlist = NULL;
	for (i = 0; i < MAX_LUNS; i++) {
		if (type == IOGENERIC)
			sprintf(device, "/dev/sg%c", ((int) ('a') + i));
		else if (type == IOFILE)
			sprintf(device, "scsi_disk_file_0_%d", i);
		else {
			printk("build_filp_table: type unknown\n");
			return -1;
		}
		printk("opening device %s\n", device);
		old_fs = get_fs();
		set_fs(get_ds());
		tmp = getname(device);
		set_fs(old_fs);
		error = PTR_ERR(tmp);
		if (IS_ERR(tmp)) {
			printk("scsi_target_init_module: getname returned an error %d\n",
				   error);
			return -1;
		}

		/* open a scsi_generic device */
# ifdef GENERICIO
		dev_file = filp_open(tmp, O_RDWR | O_NONBLOCK, 0600);
# endif

# ifdef FILEIO
		dev_file = filp_open(tmp, O_RDWR | O_NONBLOCK | O_CREAT, 0600);
# endif

		putname(tmp);
		if (IS_ERR(dev_file)) {
			error = PTR_ERR(dev_file);
			if (!i) {
				printk
					("scsi_target_init_module: filp_open returned an error %d\n",
					 error);
				return -1;
			} else
				goto out_of_loop;
		} else {
			/*
			 * add this device to the list of open devices
			 */
			temp = (devset_t *) kmalloc(sizeof (devset_t), GFP_KERNEL);
			if (!temp) {
				printk("sc..init: temp: kmalloc failed\n");
				return -1;
			}

			temp->next = NULL;
			temp->device = dev_file;
			temp->target_id = 0;
			temp->lun = i;
			temp->blk_size = BLOCKSIZE;

			printk("devices added: %s id: 0 lun: %d\n", device, i);
			if (target_data.fdlist)
				temp->next = target_data.fdlist;
			target_data.fdlist = temp;
		}
	}

  out_of_loop:
# ifdef GENERICIO
	/* also we spawn a second thread to receive signals */
	init_MUTEX_LOCKED(&target_data.signal_sem);
	init_MUTEX_LOCKED(&target_data.sig_thr_sem);
	kernel_thread((int (*)(void *)) signal_process_thread, NULL, 0);
# endif
	return 0;
}

/*
 * close_filp_table: to close all the open file descriptors
 * INPUT: None
 * OUTPUT: None
 */
static void
close_filp_table()
{
	devset_t *temp;

	while (target_data.fdlist) {
		temp = target_data.fdlist;
		target_data.fdlist = target_data.fdlist->next;
		filp_close(temp->device, NULL);
		kfree(temp);
		temp = NULL;
	}
}

# endif

/*
 * : allocates scatter-gather buffers for the received command
 * The assumption is that all front-ends use scatter-gather and so do
 * the back-ends ... this may change.
 * INPUT: Scsi_Request for which space has to be allocated, space needed
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
get_space(Scsi_Request * req, __u32 space /* in bytes */ )
{
	/* We assume that scatter gather is used universally */
	struct scatterlist *st_buffer;
	int buff_needed, i;
	__u32 count;

	/* we assume that all buffers are split by page size */

	/* get enough scatter gather entries */
	buff_needed = space / PAGE_SIZE;
	if (space > (buff_needed * PAGE_SIZE))
		buff_needed++;

	/*ramesh - added check for allocating memory */
	if (buff_needed == 0)
		buff_needed = 1;

	st_buffer =
		(struct scatterlist *) kmalloc(buff_needed *
									   sizeof (struct scatterlist),
									   GFP_KERNEL | GFP_ATOMIC);
	if (!st_buffer) {
		printk("get_space: no space for st_buffer\n");
		return -1;
	}
	memset(st_buffer, 0, buff_needed * sizeof (struct scatterlist));
	
	/* get necessary buffer space */
	for (i = 0, count = space; i < buff_needed; i++, count -= PAGE_SIZE) {
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		st_buffer[i].page = alloc_pages(GFP_DMA | GFP_KERNEL, 0);

		if (!st_buffer[i].page) {
			printk("get_space: no space for st_buffer[%d].page\n", i);
			return -1;
		}
		sg_dma_address(st_buffer + i) = (dma_addr_t)kmap(st_buffer[i].page);
		/* this should always be 0 */
		st_buffer[i].offset = (unsigned long)
			sg_dma_address(st_buffer + i) & ~PAGE_MASK;
		if (count > PAGE_SIZE)
			sg_dma_len(st_buffer + i) = PAGE_SIZE;
		else
			sg_dma_len(st_buffer + i) = count;
#else
		st_buffer[i].address =
			(__u8 *) __get_free_pages(GFP_DMA | GFP_KERNEL, 0);

		if (!st_buffer[i].address) {
			printk("get_space: no space for st_buffer[%d].address\n", i);
			return -1;
		}
		if (count > PAGE_SIZE)
			st_buffer[i].length = PAGE_SIZE;
		else
			st_buffer[i].length = count;
#endif


/* df revised begin */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,13)
/* Ming Zhang, mingz@ele.uri.edu */
#ifndef K26
		/* From 2.4.13, the structure scatterlist add a field "page", which is checked
		 *   in pci_map_sg(pci.h), which is also changed from 2.4.13. So should set a NULL as
		 *   initial value
		 */
		st_buffer[i].page = NULL;
#endif
#endif
/* df revised end */

# ifdef DEBUG_HANDLE_CMD
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		printk("st_buffer[%d] = %d\n", i, sg_dma_len(st_buffer + i));
#else
		printk("st_buffer[%d] = %d\n", i, st_buffer[i].length);
#endif
# endif
	}

	req->sr_bufflen = space;
	req->sr_buffer = st_buffer;
	req->sr_sglist_len = buff_needed * sizeof (struct scatterlist);
	req->sr_use_sg = buff_needed;

	return 0;
}

#ifdef TRUST_CDB
/*
 * get_allocation_length: This function looks at the received command
 * and calculates the size of the buffer to be allocated in order to
 * execute the command.
 * INPUT: pointer to command received
 * OUTPUT: buffer needing to be allocated or < 0 if there is an error
 */
static __u32 __attribute__ ((no_instrument_function))
get_allocation_length(__u8 *cmd)
{
	__u32 err = 0;

	switch (cmd[0]) {
	case INQUIRY:
	case MODE_SENSE:
	case MODE_SELECT:
		{
			err = cmd[ALLOC_LEN_6];
# ifdef DEBUG_ALLOCN_LEN
			printk
				("get_allocation_length: INQUIRY/MODE SENSE/MODE SELECT length %d\n",
				 err);
# endif
			break;
		}

	case WRITE_10:
	case READ_10:
	case VERIFY:
		{
			err = (cmd[ALLOC_LEN_10] << BYTE) + cmd[ALLOC_LEN_10 + 1];
			err *= BLOCKSIZE;
# ifdef DEBUG_ALLOCN_LEN
			printk("get_allocation_length: READ_10/WRITE_10 length %d\n", err);
# endif
			break;
		}

		/* cdeng, August 24 2002, Report luns */
	case REPORT_LUNS:
		{
			err = (cmd[6] << 24) + (cmd[7] << 16) + (cmd[8] << 8) + cmd[9];

# ifdef DEBUG_ALLOCN_LEN
			printk("get_allocation_length: REPORT_LUNS length %d\n", err);
# endif
			break;
		}

# ifdef TAPE_DEVICE
	case READ_6:
	case WRITE_6:
		{
			err = (cmd[2] << 16) + (cmd[3] << 8) + cmd[4];
			if (1 == cmd[1]) {
				err *= BLOCKSIZE;	/*ramesh - need to check the block size, right now it is fixed */
			}
# ifdef DEBUG_ALLOCN_LEN
			printk("get_allocation_length: READ_6/WRITE_6 length %d\n", err);
# endif
			break;
		}
#else
	case READ_6:
	case WRITE_6:
		{
			err = cmd[4] * BLOCKSIZE;
			if (err == 0)
				err = 256 * BLOCKSIZE;
# ifdef DEBUG_ALLOCN_LEN
			printk("get_allocation_length: READ_6/WRITE_6 length %d\n", err);
# endif
			break;
		}
# endif

	default:
		{
			printk("get_allocation_length: unknown command to be handled\n");
			break;
		}
	}

	return err;
}
#endif							/*ifdef TRUST_CDB */

# if defined (FILEIO) || defined (MEMORYIO)
/* 
 * get_inquiry_response: This function fills in the buffer to respond to
 * a received SCSI INQUIRY. This function is relevant when the emulator
 * is responding to commands itself and not transmitting it onto a SCSI
 * HBA attached to the system
 * INPUT: Scsi_Request ptr to the inquiry, length of expected response
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
get_inquiry_response(Scsi_Request * req, __u32 len)
{
	__u8 *inq;
	__u8 *ptr, local_buffer[36];

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	__u8 *buffer = (void *)sg_dma_address((struct scatterlist *) req->sr_buffer);
#else
	__u8 *buffer = ((struct scatterlist *) req->sr_buffer)->address;
#endif

	/* SPC-2 says in section 7.3.2 on page 82:
	 *  "The standard INQUIRY data shall contain at least 36 bytes."
	 */
	if (len >= 36) {
		/* caller's buffer is big enough to hold all 36 bytes */
		ptr = buffer;
		memset(buffer, 0x00, len);
	} else {
		/* fill a local 36-byte buffer */
		ptr = local_buffer;
		memset(local_buffer, 0x00, 36);
	}
	ptr[0] = TYPE_DISK;
	ptr[2] = 4;		/* Device complies to this standard - SPC-2 */
	ptr[3] = 2;		/* data in format specified in this standard */
	ptr[4] = 31;		/* n - 4 = 35 - 4 = 31 for full 36 byte data */
	ptr[6] = 0x80;		/* BQue = 1 basic task management supported */

	/* SPC-2 says in section 7.3.2 on page 86:
	 * "ASCII data fields shall contain only graphic codes (i.e., code
	 * values 20h through 7Eh). Left-aligned fields shall place any
	 * unused bytes at the end of the field (highest offset) and the
	 * unused bytes shall be filled with space characters (20h)."
	 */

	/* 8 byte ASCII Vendor Identification of the target - left aligned */
	memcpy(&ptr[8], "UNH-IOL ", 8);

	/* 16 byte ASCII Product Identification of the target - left aligned */
# if defined (FILEIO)
	memcpy(&ptr[16], "file-mode target", 16);
# else
	memcpy(&ptr[16], "in-memory target", 16);
#endif

	/* 4 byte ASCII Product Revision Level of the target - left aligned */
	memcpy(&ptr[32], "1.2 ", 4);

	if (len < 36) {
		/* fill as much as possible of caller's too-small buffer */
		memcpy(buffer, ptr, len);
	}

	req->sr_result = DID_OK << 16;

	if (req->sr_allowed == 1) {

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
		inq = (void *)sg_dma_address((struct scatterlist *) req->sr_buffer);
#else
		inq = ((struct scatterlist *) req->sr_buffer)->address;
#endif
		inq[0] = 0x7f;
	}

	return 0;
}

/*
 * get_read_capacity_response: This function fills up the buffer to
 * respond to a received READ_CAPACITY. This function is relevant when
 * the emulator is responding to the commands.
 * INPUT: Scsi_Request pointer to the READ_CAPACITY
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
get_read_capacity_response(Scsi_Request * req)
{
	__u32 blocksize = BLOCKSIZE, nblocks = FILESIZE / BLOCKSIZE;

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	__u8 *buffer = (void *)sg_dma_address((struct scatterlist *) req->sr_buffer);
#else
	__u8 *buffer = ((struct scatterlist *) req->sr_buffer)->address;
#endif

	memset(buffer, 0x00, READ_CAP_LEN);

	/* last block on the disk is (nblocks-1) */
	buffer[0] = ((nblocks - 1) >> (BYTE * 3)) & 0xFF;
	buffer[1] = ((nblocks - 1) >> (BYTE * 2)) & 0xFF;
	buffer[2] = ((nblocks - 1) >> (BYTE)) & 0xFF;
	buffer[3] = (nblocks - 1) & 0xFF;

	buffer[4] = (blocksize >> (BYTE * 3)) & 0xFF;
	buffer[5] = (blocksize >> (BYTE * 2)) & 0xFF;
	buffer[6] = (blocksize >> BYTE) & 0xFF;
	buffer[7] = blocksize & 0xFF;

	req->sr_result = DID_OK << 16;
	return 0;
}

/* cdeng August 24 2002
 * get_mode_sense_response: This function fills up the buffer to
 * respond to a received MODE_SENSE. This function is relevant when
 * the emulator is responding to the commands.
 * INPUT: Scsi_Request pointer to the MODE_SENSE
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
get_mode_sense_response(Scsi_Request * req, __u32 len)
{
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	__u8 *buffer = (void *)sg_dma_address((struct scatterlist *) req->sr_buffer);
#else
	__u8 *buffer = ((struct scatterlist *) req->sr_buffer)->address;
#endif

	memset(buffer, 0x00, len);

	buffer[0] = 0x0b;			// no. of bytes that follow == 11

	buffer[3] = 0x08;			// block descriptor length

	buffer[10] = 0x02;			// density code and block length

	req->sr_result = DID_OK << 16;
	return 0;
}

/* cdeng August 24 2002
 * get_report_luns_response: This function fills up the buffer to
 * respond to a received REPORT_LUNS. This function is relevant when
 * the emulator is responding to the commands.
 * INPUT: Scsi_Request pointer to the REPORT_LUNS
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
get_report_luns_response(Scsi_Request * req, __u32 len)
{
	int luns, i;

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	__u8 *buffer = (void *)sg_dma_address((struct scatterlist *) req->sr_buffer);
#else
	__u8 *buffer = ((struct scatterlist *) req->sr_buffer)->address;
#endif

	memset(buffer, 0x00, len);

	luns = (len >= (MAX_LUNS * 8 + 8)) ? MAX_LUNS : (len - 8) / 8;

	/* lun list length, MAX_LUNS needs to be less than 256 */
	buffer[2] = (luns * 8) >> 8;
	buffer[3] = (luns * 8);

	/* SAM-2, section 4.12.2 LUN 0 address
	 * "To address the LUN 0 of a SCSI device the peripheral device
	 * address method shall be used."
	 *
	 * What this means is that whenever the LUN is 0, the full 8 bytes of
	 * the LUN field in the iscsi PDU will also be 0. Because we have
	 * already zeroed out the entire buffer above,
	 * we don't need to do anything else here for lun 0.
	 */
	for (i = 1; i < luns; i++) {
		pack_lun(i, 0, buffer + i * 8 + 8);
	}

	req->sr_result = DID_OK << 16;
	return 0;
}
#endif /* defined (FILEIO) || defined (MEMORYIO) */

/*
 * hand_to_front_end: the scsi command has been tackled by the mid-level
 * and is now ready to be handed off to the front-end either because it
 * is done or because data is needed to process this command. 
 * This function is separated only because it makes 
 * the main code a lot cleaner.
 * INPUT: Scsi_Cmnd that needs to be handed off
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
hand_to_front_end(Target_Scsi_Cmnd * the_command)
{
	Scsi_Target_Device *curr_device;
# ifdef GENERICIO
	__u8 *temp;
# endif

	/* get the device template corresponding to the device_id */
	for (curr_device = target_data.st_device_list; curr_device != NULL;
		 curr_device = curr_device->next) {
		if (curr_device->id == the_command->dev_id)
			break;
	}

	if (curr_device == NULL) {	/* huh - should not get here */
		/*
		 * That there is not a device with this id may mean
		 * one of two things - there has been a mistake or
		 * that the device is no longer there. For now, the
		 * former is more probable
		 */
		printk
			("hand_to_front_end: Could not find a device with the id matching that specified in the command (%d)\n",
			 (int) the_command->dev_id);
		return -1;
	}

	/*
	 * In the time that the command was processed, it could have
	 * been aborted/released. We need to check this. If so, then
	 * the command state is changed to ST_DEQUEUE and returned
	 */
	if (the_command->abort_code != CMND_OPEN) {
		the_command->state = ST_DEQUEUE;
		return 0;
	}

	if (the_command->state == ST_DONE) {
# ifdef DEBUG_SCSI_THREAD
		printk("hand_to_front_end: to call xmit_response for %p id: %d\n",
			   the_command, the_command->id);
# endif
		if ((curr_device->template) && (curr_device->template->xmit_response)) {
			/*
			 * if it is generic then we need to get the
			 * blocksize information from the response to
			 * the READ_CAPACITY
			 */
# ifdef GENERICIO
			switch (the_command->cmd[0]) {
			case READ_CAPACITY:
				{
					temp =
						(__u8 *) ((struct scatterlist *) the_command->
										   req->sr_buffer)->address;
					the_command->fd->blk_size =
						(temp[4] << (BYTE * 3)) + (temp[5] << (BYTE * 2)) +
						(temp[6] << (BYTE)) + temp[7];
//                  printk ("READ CAPACITY: %d\n", the_command->fd->blk_size);
					break;
				}

			default:
				break;
			}
# endif

			if (curr_device->template->xmit_response(the_command)) {
				printk("hand_to_front_end: xmit_response error\n");
				return -1;
			}
			the_command->state = ST_HANDED;
		} else {
			printk
				("hand_to_front_end: no xmit_response function in the template\n");
			return -1;
		}
	} else if (the_command->state == ST_PENDING) {
# ifdef DEBUG_SCSI_THREAD
		printk("hand_to_front_end: to call rdy_to_xfer for %p id: %d\n",
			   the_command, the_command->id);
# endif
		if ((curr_device->template) && (curr_device->template->rdy_to_xfer)) {
			the_command->state = ST_XFERRED;
			if (curr_device->template->rdy_to_xfer(the_command)) {
				printk("hand_to_front_end: rdy_to_xfer error\n");
				return -1;
			}
		} else {
			printk
				("hand_to_front_end: no rdy_to_xfer function in the template\n");
			return -1;
		}
	} else {
		printk
			("hand_to_front_end: command %p id: %d state: %d should not have reached this function\n",
			 the_command, the_command->id, the_command->state);
		return -1;
	}

	return 0;
}

/*
 * abort_notify: This function is used to notify to the front end driver
 * that a command has been successfully aborted
 * INPUT: Target_Scsi_Cmnd
 * OUTPUT: none
 */
static int
abort_notify(Target_Scsi_Message * msg)
{
	Target_Scsi_Cmnd *cmnd;
	Scsi_Target_Device *curr_device;

	if (msg && msg->value)
		cmnd = (Target_Scsi_Cmnd *) msg->value;
	else {
		printk("abort_notify: null cmnd in the msg\n");
		return -1;
	}

	/* get the device template corresponding to the device_id */
	for (curr_device = target_data.st_device_list; curr_device != NULL;
		 curr_device = curr_device->next) {
		if (curr_device->id == cmnd->dev_id)
			break;
	}

	if (curr_device == NULL) {
		printk("abort_notify: Could not find the device\n");
		return -1;
	}

	if ((curr_device->template) && (curr_device->template->task_mgmt_fn_done))
		curr_device->template->task_mgmt_fn_done(msg);
	else {
		printk
			("abort_notify: Unable to notify front end about abort notification\n");
		return -1;
	}

	return 0;
}

/*
 * aen_notify: This function is used to notify to the low-level driver
 * that task management function has been received and that they need to
 * contact all initiators logged in to them and let them know. NOTE: the
 * aen_notify will also call the device that sent the task management
 * function - it needs to inform other initiators (than the one that
 * originated the Task Management function) about the received Mgmt fn
 * INPUT: management function and lun
 * OUTPUT: None
 */
static void
aen_notify(int fn, __u64 lun)
{
	Scsi_Target_Device *dev;

	for (dev = target_data.st_device_list; dev != NULL; dev = dev->next) {
		if (dev && dev->template && dev->template->report_aen)
			dev->template->report_aen(fn, lun);
		else
			printk("aen_notify: Unable to notify device %d\n", (int) dev->id);
	}
}

/********************************************************************
 * THESE ARE FUNCTIONS WHICH ARE SPECIFIC TO MEMORY IO - I lump them*
 * 	together just to help me keep track of what is where	    *
 *******************************************************************/

# ifdef MEMORYIO
/*
 * handle_cmd: This command is responsible for dealing with the received
 * command. I am slightly unclear about everything this function needs
 * to do. Right now, it receives new commands, and gets them to the
 * state where they are processing or pending.
 * INPUT: Target_Scsi_Cmnd to be handled
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
handle_cmd(Target_Scsi_Cmnd * cmnd)
{
	int err = 0;
	__u32 to_read;

# ifdef DEBUG_HANDLE_CMD
	printk("Entering handle_cmd\n");
# endif
	switch (cmnd->req->sr_cmnd[0]) {

	case READ_CAPACITY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_CAPACITY received\n");
# endif
			/* perform checks on READ_CAPACITY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* allocate sg_list and get_free_pages */
			if (get_space(cmnd->req, READ_CAP_LEN)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* get response */
			if (get_read_capacity_response(cmnd->req)) {
				printk
					("handle_command: get_read_capacity_response returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;

			err = 0;
			break;
		}

	case INQUIRY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("INQUIRY received\n");
# endif
			/* perform checks on INQUIRY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* get response */
			if (get_inquiry_response(cmnd->req, to_read)) {
				printk
					("handle_command: get_inquiry_response returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;

			err = 0;
			break;
		}

	case TEST_UNIT_READY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("TEST UNIT READY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->req->sr_result = DID_OK << 16;
			cmnd->state = ST_DONE;
			err = 0;
			break;
		}

	case REPORT_LUNS:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("Report luns command received\n");
# endif
			/* perform checks on Report LUNS - LATER */
			if (cmnd->req->sr_cmnd[2] != 0) {
				printk
					("handle_cmnd: the Select_Report in report_luns is not zero\n");
			}

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* get response */
			if (get_report_luns_response(cmnd->req, to_read)) {
				printk
					("handle_command: get_report_luns_response returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

	case MODE_SENSE:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("MODE_SENSE received\n");
# endif
			/* perform checks on MODE_SENSE - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* get response */
			if (get_mode_sense_response(cmnd->req, to_read)) {
				printk
					("handle_command: get_mode_sense_response returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

	case VERIFY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("VERIFY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

	case READ_6:
	case READ_10:
		{
# ifdef DEBUG_HANDLE_CMD
			if (cmnd->req->sr_cmnd[0] == READ_6)
				printk("READ_6 received\n");
			else
				printk("READ_10 received\n");
# endif
			/* perform checks for READ_10 */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;
			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_cmd: get_allocation_length returned (%d) an error for %d\n",
					 to_read, cmnd->id);
				err = -1;
				break;
			}

			if (get_space(cmnd->req, to_read)) {
				printk("handle_cmd: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* this data can just be returned from memory */
			/* change_status */
			cmnd->req->sr_result = DID_OK << 16;
			cmnd->state = ST_DONE;
			err = 0;
			break;
		}

	case WRITE_6:
	case WRITE_10:
		{
# ifdef DEBUG_HANDLE_CMD
			if (cmnd->req->sr_cmnd[0] == WRITE_6)
				printk("WRITE_6 received\n");
			else
				printk("WRITE_10 received\n");
# endif
			if (cmnd->state == ST_NEW_CMND) {
				/* perform checks on the received WRITE_10 */

				/* set data direction for the WRITE_10 */
				cmnd->req->sr_data_direction = SCSI_DATA_WRITE;
				/* get length */
				to_read = get_allocation_length(cmnd->req->sr_cmnd);

				if (to_read < 0) {
					printk
						("handle_cmd: get_allocation_length returned (%d) an error for %d\n",
						 to_read, cmnd->id);
					err = -1;
					break;
				}

				/* get space */
				if (get_space(cmnd->req, to_read)) {
					printk("handle_cmd: get_space returned error for %d\n",
						   cmnd->id);
					err = -1;
					break;
				}
				cmnd->state = ST_PENDING;
			} else if (cmnd->state == ST_TO_PROCESS) {
				/* 
				 * in memory mode of processing we do not
				 * care about the data received at all
				 */
				cmnd->state = ST_DONE;
				cmnd->req->sr_result = DID_OK << 16;
			}
			err = 0;
			break;
		}

	default:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("unknown command received\n");
# endif
			break;
		}
	}
	return err;
}

#endif

/********************************************************************
 * THESE ARE FUNCTIONS WHICH ARE SPECIFIC TO DISK IO - I lump them  *
 * 	together just to help me keep track of what is where	    *
 *******************************************************************/

# ifdef DISKIO

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26

/* Need to add EXPORT_SYMBOL(scsi_bus_type) */
extern struct bus_type scsi_bus_type; 

static int fill_scsidevice(struct device *dev, void *data)
{
	Scsi_Device *the_device = to_scsi_device(dev);
	int i;

	printk("device type - %d ", the_device->type);
# ifdef TAPE_DEVICE
	if (the_device->type == TYPE_TAPE)
# else
	if (the_device->type == TYPE_DISK)
# endif
	{
		for (i = 0; i < target_num; i++) {
			if (target_info[i][0] == the_device->host->host_no &&
				target_info[i][1] == the_device->channel &&
				target_info[i][2] == the_device->id) {
				scsi_device[i][the_device->lun] = the_device;
				break;
			}
		}

		if (i >= target_num) {
			target_info[target_num][0] = the_device->host->host_no;
			target_info[target_num][1] = the_device->channel;
			target_info[target_num][2] = the_device->id;
			scsi_device[target_num][the_device->lun] = the_device;
			target_num++;
		}
	}
		
	return 0;
}

/*
 * fill_scsi_device: This function is responsible for filling in the
 * details of the device that will be used by the Target Emulator to
 * deal with the SCSI commands. If no SCSI devices are found attached
 * to the system then an error is returned
 * INPUT: NONE
 * OUTPUT: Scsi_Device* if okay, NULL if there is trouble
 */
static int
fill_scsi_device(void)
{
	target_num = 0;
	// scan each device	
	bus_for_each_dev(&scsi_bus_type, NULL, NULL, fill_scsidevice);

	if (target_num == 0) {
		printk
			("fill_scsi_device: could not find a SCSI device on the system .. this mode of operation is doomed for failure bailing out\n");
		return -1;
	} else
		printk("fill_scsi_device: %d devices detected\n", target_num);

	return 0;
}

#else

/*
 * fill_scsi_device: This function is responsible for filling in the
 * details of the device that will be used by the Target Emulator to
 * deal with the SCSI commands. If no SCSI devices are found attached
 * to the system then an error is returned
 * INPUT: NONE
 * OUTPUT: Scsi_Device* if okay, NULL if there is trouble
 */
static int
fill_scsi_device(void)
{
	struct Scsi_Host *the_host = scsi_hostlist;
	Scsi_Device *the_device = NULL;
	int i;

	/* check out the first HBA */
	while (the_host) {
		the_device = the_host->host_queue;
		/* check out the first device on this HBA */
		while (the_device) {
			printk("device type - %d ", the_device->type);

# ifdef TAPE_DEVICE
			if (the_device->type == TYPE_TAPE)
# else
			if (the_device->type == TYPE_DISK)
# endif
			{
				printk("Host: %d Channel: %d ID: %d LUN: %d\n",
					   the_device->host->host_no, the_device->channel,
					   the_device->id, the_device->lun);

				for (i = 0; i < target_num; i++) {
					if (target_info[i][0] == the_device->host->host_no &&
						target_info[i][1] == the_device->channel &&
						target_info[i][2] == the_device->id) {
						scsi_device[i][the_device->lun] = the_device;
						break;
					}
				}

				if (i >= target_num) {
					target_info[target_num][0] = the_device->host->host_no;
					target_info[target_num][1] = the_device->channel;
					target_info[target_num][2] = the_device->id;
					scsi_device[target_num][the_device->lun] = the_device;
					target_num++;
				}
			}

			the_device = the_device->next;

		}

		the_host = the_host->next;
	}

	if (target_num == 0) {
		printk
			("fill_scsi_device: could not find a SCSI device on the system .. this mode of operation is doomed for failure bailing out\n");
		return -1;
	} else
		printk("fill_scsi_device: %d devices detected\n", target_num);

	return 0;
}

#endif /* K26 */

/* if we trust the information of iSCSI from initiator, then
 * set the req and malloc space using that information, else
 * we parse the cdb to get correct information ourselves */
#ifndef TRUST_CDB

/* 
 * handle_cmd: This command is responsible for dealing with the received
 * command when a real is used by the scsi target emulator. It receives
 * new commands, and gets them to the state where they are processing or
 * pending.
 * INPUT: Target_Scsi_Cmnd to be handled
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 * REMARKS - ramesh@global.com : changed the code to not to parse the
 *                               input cdb. this makes the code to be
 *                               independent of devices.
 */
static int
handle_cmd(Target_Scsi_Cmnd * cmnd)
{
	int err = -1;
	int allocate_buff = 1;

# ifdef DEBUG_HANDLE_CMD
	printk("Entering handle_cmd\n");
# endif

	/* set data direction */
	cmnd->req->sr_data_direction = SCSI_DATA_NONE;
	if (cmnd->flags & R_BIT) {
		cmnd->req->sr_data_direction = SCSI_DATA_READ;
	} else if (cmnd->flags & W_BIT) {
		cmnd->req->sr_data_direction = SCSI_DATA_WRITE;
	}
# ifdef DEBUG_HANDLE_CMD
	printk("handle_cmd: data direction %d data len %d\n",
		   cmnd->req->sr_data_direction, cmnd->datalen);
# endif

	/*check the data length need to be allocated */
	if (cmnd->datalen) {
		if (cmnd->req->sr_data_direction == SCSI_DATA_WRITE &&
			cmnd->state == ST_TO_PROCESS) {
			allocate_buff = 0;
		}

		if (allocate_buff) {
			/* allocate sg_list and get_free_pages */
			if (get_space(cmnd->req, cmnd->datalen)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				return err;
			}
		}
	} else {
		cmnd->req->sr_use_sg = 0;
		cmnd->req->sr_bufflen = 0;
	}
# ifdef DEBUG_HANDLE_CMD
	printk("handle_cmd: buffer length %d\n", cmnd->req->sr_bufflen);
# endif

	/* change status */
	/* for write needs more data to arive... */
	if (cmnd->req->sr_data_direction == SCSI_DATA_WRITE &&
		cmnd->state == ST_NEW_CMND) {
		cmnd->state = ST_PENDING;
	} else {
		cmnd->state = ST_PROCESSING;
		/* hand it off to the mid-level to deal with */
		scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
					cmnd->req->sr_bufflen, te_cmnd_processed,
					TE_TIMEOUT, TE_TRY);
# ifdef DEBUG_HANDLE_CMD
		printk("handle_cmd: handing it off to mid_level\n");
# endif
	}

	return 0;
}

#else							/* TRUST_CDB */

static int
handle_cmd(Target_Scsi_Cmnd * cmnd)
{
	int err = 0;
	__u32 to_read;

# ifdef DEBUG_HANDLE_CMD
	printk("Entering handle_cmd\n");
# endif

	switch (cmnd->req->sr_cmnd[0]) {
	case READ_CAPACITY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_CAPACITY received\n");
# endif
			/* perform checks on READ_CAPACITY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* allocate sg_list and get_free_pages */
			if (get_space(cmnd->req, READ_CAP_LEN)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
						cmnd->req->sr_bufflen, te_cmnd_processed, TE_TIMEOUT,
						TE_TRY);

			err = 0;
			break;
		}

	case INQUIRY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("INQUIRY received\n");
# endif
			/* perform checks on INQUIRY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);
			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
						cmnd->req->sr_bufflen, te_cmnd_processed, TE_TIMEOUT,
						TE_TRY);

			err = 0;
			break;
		}

	case REPORT_LUNS:			// cdeng        July 29, 2002
		{
# ifdef DEBUG_HANDLE_CMD
			printk("REPORT_LUNS received\n");
# endif
			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
						cmnd->req->sr_bufflen, te_cmnd_processed, TE_TIMEOUT,
						TE_TRY);

			err = 0;
			break;
		}

	case MODE_SENSE:			// cdeng    May 28, 2002
		{
# ifdef DEBUG_HANDLE_CMD
			printk("MODE_SENSE received\n");
# endif
			/* perform checks on MODE_SENSE - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
						cmnd->req->sr_bufflen, te_cmnd_processed, TE_TIMEOUT,
						TE_TRY);

			err = 0;
			break;
		}

	case TEST_UNIT_READY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("TEST UNIT READY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
						cmnd->req->sr_bufflen, te_cmnd_processed, TE_TIMEOUT,
						TE_TRY);

			err = 0;
			break;
		}

	case VERIFY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("VERIFY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
						cmnd->req->sr_bufflen, te_cmnd_processed, TE_TIMEOUT,
						TE_TRY);

			err = 0;
			break;
		}

	case READ_10:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_10 received\n");
# endif
			/* perform checks for READ_10 */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;
			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_cmd: get_allocation_length returned (%d) an error for %d\n",
					 to_read, cmnd->id);
				err = -1;
				break;
			}

			if (get_space(cmnd->req, to_read)) {
				printk("handle_cmd: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* change_status */
			cmnd->state = ST_PROCESSING;

			/* hand it off to the mid-level to deal with */
			scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
						cmnd->req->sr_bufflen, te_cmnd_processed, TE_TIMEOUT,
						TE_TRY);

			err = 0;
			break;
		}

	case WRITE_10:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("WRITE_10 received\n");
# endif
			if (cmnd->state == ST_NEW_CMND) {
				/* perform checks on the received WRITE_10 */

				/* set data direction for the WRITE_10 */
				cmnd->req->sr_data_direction = SCSI_DATA_WRITE;
				/* get length */
				to_read = get_allocation_length(cmnd->req->sr_cmnd);

				if (to_read < 0) {
					printk
						("handle_cmd: get_allocation_length returned (%d) an error for %d\n",
						 to_read, cmnd->id);
					err = -1;
					break;
				}

				/* get space */
				if (get_space(cmnd->req, to_read)) {
					printk("handle_cmd: get_space returned error for %d\n",
						   cmnd->id);
					err = -1;
					break;
				}
				cmnd->state = ST_PENDING;
			} else if (cmnd->state == ST_TO_PROCESS) {
				cmnd->state = ST_PROCESSING;

				/* hand it off to the mid-level to deal with */
				scsi_do_req(cmnd->req, cmnd->cmd, cmnd->req->sr_buffer,
							cmnd->req->sr_bufflen, te_cmnd_processed,
							TE_TIMEOUT, TE_TRY);
			}
			err = 0;
			break;
		}

	default:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("unknown command received\n");
# endif
			err = -1;
			break;
		}
	}
# ifdef DEBUG_HANDLE_CMD
	printk("handle_cmd: handing it off to mid_level\n");
# endif
	return err;
}
#endif							/* end of TRUST_CDB */

/*
 * te_cmnd_processed: This is the done function called by the mid-level
 * upon detecting that a particular command has been executed. This
 * function can be called from an interrupt context - it is responsible
 * for changing the state of the command and waking up the target 
 * mid-level.
 */
static void
te_cmnd_processed(Scsi_Cmnd * sc_cmd)
{
	Scsi_Request *req;
	Target_Scsi_Cmnd *te_cmd;

	/* find the scsi_request */
	req = sc_cmd->sc_request;

	/* find the Target_Scsi_Cmnd containing the scsi_request */
	te_cmd = target_data.cmd_queue_start;
	while (te_cmd) {
		if (te_cmd->req == req)
			break;
		te_cmd = te_cmd->next;
	}

	if (!te_cmd) {				/* we can just throw away the response */
		/* may be the command was aborted */
		printk
			("te_cmnd_processed: Could not find request in the command queue\n");
		scsi_release_request(req);
		req = NULL;
	} else {
		/* change its state */
# ifdef DEBUG_HANDLE_CMD
		printk("Okay obtained response for cmd: %p id: %d \n", te_cmd,
			   te_cmd->id);
# endif

		/* if the lun of device is not the lun of request */
		if (te_cmd->lun != req->sr_device->lun && req->sr_cmnd[0] == INQUIRY) {
			__u8 *inq;

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			inq = page_address(((struct scatterlist *) req->sr_buffer)->page);
			if (inq) {
				inq[0] = 0x7f;	// 011 + 1FH
			}
#else
			inq = ((struct scatterlist *) req->sr_buffer)->address;
			inq[0] = 0x7f;
#endif
		}

		te_cmd->state = ST_DONE;
		up(&target_data.target_sem);
	}

	return;
}

# endif

/***********************************************************************
 * THESE ARE FUNCTIONS WHICH ARE SPECIFIC TO GENERIC IO - I lump them  *
 *      together just to help me keep track of what is where           *
 ***********************************************************************/

# ifdef GENERICIO
/*
 * handle_cmd: This command is responsible for dealing with the received
 * command. I am slightly unclear about everything this function needs
 * to do. Right now, it receives new commands, and gets them to the
 * state where they are processing or pending.
 * INPUT: Target_Scsi_Cmnd to be handled
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
handle_cmd(Target_Scsi_Cmnd * cmnd)
{
	int err = 0;
	mm_segment_t old_fs;
	struct file *dev_file;
	__u32 to_read;

# ifdef DEBUG_HANDLE_CMD
	printk("Entering handle_cmd\n");
# endif

	if (cmnd->fd)
		dev_file = cmnd->fd->device;
	else {
		printk("handle_cmd: No file descriptor available\n");
		return -1;
	}

	switch (cmnd->req->sr_cmnd[0]) {
	case READ_CAPACITY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_CAPACITY received\n");
# endif
			/* perform checks on READ_CAPACITY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* allocate sg_list and get_free_pages */
			if (get_space(cmnd->req, READ_CAP_LEN)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			if (fill_sg_hdr(cmnd, SG_DXFER_FROM_DEV)) {
				printk("handle_cmd: fill_sg_hdr returned an error\n");
				return -1;
			}

			/* get response */
			if ((dev_file) && (dev_file->f_op) && (dev_file->f_op->write)) {
				old_fs = get_fs();
				set_fs(get_ds());
				err =
					dev_file->f_op->write(dev_file, (__u8 *) cmnd->sg,
										  sizeof (sg_io_hdr_t),
										  &dev_file->f_pos);
				set_fs(old_fs);
				if (err < 0) {
					printk("handle_cmd: READ_CAPACITY: write returned %d\n",
						   err);
					break;
				}
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			err = 0;
			break;
		}

	case INQUIRY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("INQUIRY received\n");
# endif
			/* perform checks on INQUIRY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			if (fill_sg_hdr(cmnd, SG_DXFER_FROM_DEV)) {
				printk("handle_cmd: fill_sg_hdr returned an error\n");
				return -1;
			}

			/* get response */
			if ((dev_file) && (dev_file->f_op) && (dev_file->f_op->write)) {
				old_fs = get_fs();
				set_fs(get_ds());
				err =
					dev_file->f_op->write(dev_file, (__u8 *) cmnd->sg,
										  sizeof (sg_io_hdr_t),
										  &dev_file->f_pos);
				set_fs(old_fs);
				if (err < 0) {
					printk("ioctl returned %d\n", err);
					break;
				}
			}

			/* change status */
			cmnd->state = ST_PROCESSING;

			err = 0;
			break;
		}

	case TEST_UNIT_READY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("TEST UNIT READY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;

			if (fill_sg_hdr(cmnd, SG_DXFER_NONE)) {
				printk("handle_cmd: fill_sg_hdr returned an error\n");
				return -1;
			}

			/* get response */
			if ((dev_file) && (dev_file->f_op) && (dev_file->f_op->write)) {
				old_fs = get_fs();
				set_fs(get_ds());
				err =
					dev_file->f_op->write(dev_file, (__u8 *) cmnd->sg,
										  sizeof (sg_io_hdr_t),
										  &dev_file->f_pos);
				set_fs(old_fs);
				if (err < 0) {
					printk("handle_cmd: TEST UNIT READY: write returned %d\n",
						   err);
					break;
				}
			}

			cmnd->state = ST_PROCESSING;
			err = 0;
			break;
		}

	case READ_10:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_10 received\n");
# endif
			/* perform checks for READ_10 */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;
			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_cmd: get_allocation_length returned (%d) an error for %d\n",
					 to_read, cmnd->id);
				err = -1;
				break;
			}

			/* my stupidity */
			to_read = to_read / BLOCKSIZE * cmnd->fd->blk_size;
			if (get_space(cmnd->req, to_read)) {
				printk("handle_cmd: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			if (fill_sg_hdr(cmnd, SG_DXFER_FROM_DEV)) {
				printk("handle_cmd: fill_sg_hdr returned an error\n");
				return -1;
			}

			/* get response */
			if ((dev_file) && (dev_file->f_op) && (dev_file->f_op->write)) {
				old_fs = get_fs();
				set_fs(get_ds());
				err =
					dev_file->f_op->write(dev_file, (__u8 *) cmnd->sg,
										  sizeof (sg_io_hdr_t),
										  &dev_file->f_pos);
				set_fs(old_fs);
				if (err < 0) {
					printk("handle_cmd: READ_CAPACITY: write returned %d\n",
						   err);
					break;
				}
			}

			/* change_status */
			cmnd->state = ST_PROCESSING;
			err = 0;
			break;
		}

	case WRITE_10:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("WRITE_10 received\n");
# endif
			if (cmnd->state == ST_NEW_CMND) {
				/* perform checks on the received WRITE_10 */

				/* set data direction for the WRITE_10 */
				cmnd->req->sr_data_direction = SCSI_DATA_WRITE;
				/* get length */
				to_read = get_allocation_length(cmnd->req->sr_cmnd);

				if (to_read < 0) {
					printk
						("handle_cmd: get_allocation_length returned (%d) an error for %d\n",
						 to_read, cmnd->id);
					err = -1;
					break;
				}

				/* my stupidity */
				to_read = to_read / BLOCKSIZE * cmnd->fd->blk_size;
				/* get space */
				if (get_space(cmnd->req, to_read)) {
					printk("handle_cmd: get_space returned error for %d\n",
						   cmnd->id);
					err = -1;
					break;
				}

				cmnd->state = ST_PENDING;
			} else if (cmnd->state == ST_TO_PROCESS) {
				if (fill_sg_hdr(cmnd, SG_DXFER_TO_DEV)) {
					printk("handle_cmd: fill_sg_hdr returned an error\n");
					return -1;
				}

				/* get response */
				if ((dev_file) && (dev_file->f_op)
					&& (dev_file->f_op-> /*ioctl */ write)) {
					old_fs = get_fs();
					set_fs(get_ds());
//                  err = dev_file->f_op->ioctl (dev_file->f_dentry->d_inode, dev_file, SG_IO, (unsigned long) cmnd->sg);
					err =
						dev_file->f_op->write(dev_file,
											  (__u8 *) cmnd->sg,
											  sizeof (sg_io_hdr_t),
											  &dev_file->f_pos);
					set_fs(old_fs);
					if (err < 0) {
						printk("handle_cmd: WRITE_10: write returned %d\n",
							   err);
						break;
					}
				}

				cmnd->state = ST_PROCESSING;
//              cmnd->state = ST_DONE;
			}
			err = 0;
			break;
		}

	default:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("unknown command received\n");
# endif
			break;
		}
	}

	return err;
}

/*
 * fill_sg_hdr: This function fills in the values required by the sg
 * struct. The sg_hdr is kmalloced, filled and returned to the calling
 * function.
 * INPUT: Target_Scsi_Cmnd to process, direction of transfer (use the
 * values specified by SCSI generic
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
fill_sg_hdr(Target_Scsi_Cmnd * cmnd, int direction)
{
	int i;
	sg_iovec_t *sg_iov;
	struct scatterlist *st_list;

	cmnd->sg =
		(sg_io_hdr_t *) kmalloc(sizeof (sg_io_hdr_t), GFP_KERNEL | GFP_ATOMIC);
	if (!cmnd->sg) {
		printk("fill_sg_hdr: cmnd->sg: kmalloc failed\n");
		return -1;
	}

	cmnd->sg->interface_id = 'S';	// for SCSI generic
	cmnd->sg->dxfer_direction = direction;
	cmnd->sg->cmd_len = cmnd->len;
	cmnd->sg->mx_sb_len = MAX_SENSE_DATA;
	cmnd->sg->iovec_count = cmnd->req->sr_use_sg;
	cmnd->sg->dxfer_len = cmnd->req->sr_bufflen;
//  cmnd->sg->dxferp handled later
	cmnd->sg->cmdp = cmnd->req->sr_cmnd;
	cmnd->sg->sbp =
		(__u8 *) kmalloc(MAX_SENSE_DATA * sizeof (__u8),
								  GFP_KERNEL | GFP_ATOMIC);
	cmnd->sg->timeout = TE_TIMEOUT;
	cmnd->sg->flags = 0;

	/* we need the data buffers to get pointed correctly */
	if (cmnd->req->sr_bufflen) {	// there is data

		sg_iov =
			(sg_iovec_t *) kmalloc(cmnd->sg->iovec_count * sizeof (sg_iovec_t),
								   GFP_KERNEL | GFP_ATOMIC);
		if (!sg_iov) {
			printk("fill_sg_hdr: sg_iov: kmalloc failed\n");
			return -1;
		}

		st_list = (struct scatterlist *) cmnd->req->sr_buffer;

		for (i = 0; i < cmnd->sg->iovec_count; i++) {

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
			sg_iov[i].iov_base = (void *)sg_dma_address(st_list + i);
			sg_iov[i].iov_len = sg_dma_len(st_list + i);
#else
			sg_iov[i].iov_base = st_list[i].address;
			sg_iov[i].iov_len = st_list[i].length;
#endif
			
# ifdef DEBUG_HANDLE_CMD
//          printk ("sg_iov[%d]: %d\n", i, sg_iov[i].iov_len);
# endif
		}
		cmnd->sg->dxferp = sg_iov;
	} else
		cmnd->sg->dxferp = NULL;

	return 0;
}
# endif

/********************************************************************
 * THESE ARE FUNCTIONS WHICH ARE SPECIFIC TO FILE IO - I lump them  *
 *      together just to help me keep track of what is where        *
 *******************************************************************/

# ifdef FILEIO
/*
 * handle_cmd: This command is responsible for dealing with the received
 * command. I am slightly unclear about everything this function needs
 * to do. Right now, it receives new commands, and gets them to the
 * state where they are processing or pending.
 * INPUT: Target_Scsi_Cmnd to be handled
 * OUTPUT: 0 if everything is okay, < 0 if there is trouble
 */
static int
handle_cmd(Target_Scsi_Cmnd * cmnd)
{
	struct file *dev_file;
	int err = 0, i;
	mm_segment_t old_fs;
	__u32 to_read, lba_start;
	struct scatterlist *st_list;

# ifdef DEBUG_HANDLE_CMD
	printk("Entering handle_cmd\n");
# endif
	switch (cmnd->req->sr_cmnd[0]) {
	case READ_CAPACITY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_CAPACITY received\n");
# endif
			/* perform checks on READ_CAPACITY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* allocate sg_list and get_free_pages */
			if (get_space(cmnd->req, READ_CAP_LEN)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* get response */
			if (get_read_capacity_response(cmnd->req)) {
				printk
					("handle_command: get_read_capacity_response returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;

			err = 0;
			break;
		}

	case INQUIRY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("INQUIRY received\n");
# endif
			/* perform checks on INQUIRY - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* get response */
			if (get_inquiry_response(cmnd->req, to_read)) {
				printk
					("handle_command: get_inquiry_response returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

		/* cdeng, August 24 2002, report luns command */
	case REPORT_LUNS:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("Report luns command received\n");
# endif
			/* perform checks on Report LUNS - LATER */
			if (cmnd->req->sr_cmnd[2] != 0) {
				printk
					("handle_cmnd: the Select_Report in report_luns is not zero\n");
			}

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* get response */
			if (get_report_luns_response(cmnd->req, to_read)) {
				printk
					("handle_command: get_report_luns_response returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

		/* cdeng, August 24 2002 */
	case MODE_SENSE:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("MODE_SENSE received\n");
# endif
			/* perform checks on MODE_SENSE - LATER */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;

			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_command: get_allocation length returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* allocate space */
			if (get_space(cmnd->req, to_read)) {
				printk("handle_command: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* get response */
			if (get_mode_sense_response(cmnd->req, to_read)) {
				printk
					("handle_command: get_mode_sense_response returned an error for %d\n",
					 cmnd->id);
				err = -1;
				break;
			}

			/* change status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

		/* cdeng, August 24 2002 */
	case VERIFY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("VERIFY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

	case START_STOP:
	case TEST_UNIT_READY:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("TEST UNIT READY received\n");
# endif
			/* perform checks on TEST UNIT READY */
			cmnd->req->sr_data_direction = SCSI_DATA_NONE;
			cmnd->req->sr_use_sg = 0;
			cmnd->req->sr_bufflen = 0;
			cmnd->req->sr_result = DID_OK << 16;
			cmnd->state = ST_DONE;
			err = 0;
			break;
		}

	case READ_10:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("READ_10 received\n");
# endif
			/* perform checks for READ_10 */

			/* set data direction */
			cmnd->req->sr_data_direction = SCSI_DATA_READ;
			/* get length */
			to_read = get_allocation_length(cmnd->req->sr_cmnd);

			if (to_read < 0) {
				printk
					("handle_cmd: get_allocation_length returned (%d) an error for %d\n",
					 to_read, cmnd->id);
				err = -1;
				break;
			}

			if (get_space(cmnd->req, to_read)) {
				printk("handle_cmd: get_space returned an error for %d\n",
					   cmnd->id);
				err = -1;
				break;
			}

			/* data needs to be returned from the file */
			lba_start =
				(cmnd->req->sr_cmnd[LBA_POSN_10] << (BYTE * 3)) +
				(cmnd->req->sr_cmnd[LBA_POSN_10 + 1] << (BYTE * 2)) +
				(cmnd->req->sr_cmnd[LBA_POSN_10 + 2] << (BYTE)) +
				cmnd->req->sr_cmnd[LBA_POSN_10 + 3];

			st_list = (struct scatterlist *) cmnd->req->sr_buffer;

			dev_file = cmnd->fd->device;
			for (i = 0; i < cmnd->req->sr_use_sg; i++) {
				if (dev_file) {
					if (dev_file->f_op->llseek) {
						lock_kernel();
						old_fs = get_fs();
						set_fs(get_ds());
						err =
							dev_file->f_op->llseek(dev_file,
												   lba_start * BLOCKSIZE,
												   0 /*SEEK_SET */ );
						set_fs(old_fs);
						unlock_kernel();
					} else {
						lock_kernel();
						old_fs = get_fs();
						set_fs(get_ds());
						err =
							default_llseek(dev_file, lba_start * BLOCKSIZE,
										   0 /*SEEK_SET */ );
						set_fs(old_fs);
						unlock_kernel();
					}

					if (err != lba_start * BLOCKSIZE) {
						printk("handle_cmd: lseek trouble\n");
						break;
					}

					if ((dev_file->f_op)
						&& (dev_file->f_op->read)) {
						old_fs = get_fs();
						set_fs(get_ds());
						err = dev_file->f_op->read(
							dev_file,
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
							(void *)sg_dma_address(st_list + i),
							sg_dma_len(st_list + i),
#else
							st_list[i].address,
							st_list[i].length,
#endif
							&dev_file->f_pos);

						set_fs(old_fs);

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
						if (err < sg_dma_len(st_list + i)) {
#else
						if (err < st_list[i].length) {
#endif
							if (err < 0) {
								if (err != -EAGAIN) {
									printk("handle_cmnd: read returned %d\n",
										   err);
									break;
								}
								err = 0;
							}
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
							memset(
								(void *)sg_dma_address(st_list + i) + err,
								0,
								sg_dma_len(st_list + i) - err);
#else
							memset(
								st_list[i].address + err,
								0,
								st_list[i].length - err);
#endif
						}
//                      if (err > 0)
//                          inode_dir_notify (dev_file->f_dentry->d_parent->d_inode, DN_ACCESS);
					}
				}

				/*
				 * Assumption is that everything will be
				 * written in blocksize chunks. Also the
				 * blocksize is less than pagesize and
				 * the latter is a multiple of the first
				 */
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26 
				lba_start += sg_dma_len(st_list + i) / BLOCKSIZE;
#else
				lba_start += st_list[i].length / BLOCKSIZE;
#endif
			}
			/* change_status */
			cmnd->state = ST_DONE;
			cmnd->req->sr_result = DID_OK << 16;
			err = 0;
			break;
		}

	case WRITE_10:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("WRITE_10 received\n");
# endif
			if (cmnd->state == ST_NEW_CMND) {
				/* perform checks on the received WRITE_10 */

				/* set data direction for the WRITE_10 */
				cmnd->req->sr_data_direction = SCSI_DATA_WRITE;
				/* get length */
				to_read = get_allocation_length(cmnd->req->sr_cmnd);

				if (to_read < 0) {
					printk
						("handle_cmd: get_allocation_length returned (%d) an error for %d\n",
						 to_read, cmnd->id);
					err = -1;
					break;
				}

				/* get space */
				if (get_space(cmnd->req, to_read)) {
					printk("handle_cmd: get_space returned error for %d\n",
						   cmnd->id);
					err = -1;
					break;
				}
				cmnd->state = ST_PENDING;
			} else if (cmnd->state == ST_TO_PROCESS) {
				/* data needs to be written to the file */
				lba_start =
					(cmnd->req->sr_cmnd[LBA_POSN_10] << (BYTE * 3)) +
					(cmnd->req->sr_cmnd[LBA_POSN_10 + 1] << (BYTE * 2)) +
					(cmnd->req->sr_cmnd[LBA_POSN_10 + 2] << (BYTE)) +
					cmnd->req->sr_cmnd[LBA_POSN_10 + 3];

				st_list = (struct scatterlist *) cmnd->req->sr_buffer;

				dev_file = cmnd->fd->device;
				for (i = 0; i < cmnd->req->sr_use_sg; i++) {
					if (dev_file) {
						if (dev_file->f_op->llseek) {
							lock_kernel();
							old_fs = get_fs();
							set_fs(get_ds());
							err =
								dev_file->f_op->llseek(dev_file,
													   lba_start * BLOCKSIZE,
													   0 /*SEEK_SET */ );
							set_fs(old_fs);
							unlock_kernel();
						} else {
							lock_kernel();
							old_fs = get_fs();
							set_fs(get_ds());
							err =
								default_llseek(dev_file, lba_start * BLOCKSIZE,
											   0 /*SEEK_SET */ );
							set_fs(old_fs);
							unlock_kernel();
						}

						if (err != lba_start * BLOCKSIZE) {
							printk("handle_cmd: lseek trouble\n");
							break;
						}

						if ((dev_file->f_op)
							&& (dev_file->f_op->read)) {
							old_fs = get_fs();
							set_fs(get_ds());
							err =
							dev_file->f_op->write(
							  dev_file,
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
							  (void *)sg_dma_address(st_list + i),
							  sg_dma_len(st_list + i),
#else
							  st_list[i].address,
							  st_list[i].length,
#endif
							  &dev_file->f_pos);
							set_fs(old_fs);
							if ((err < 0) && (err != -EAGAIN)) {
								printk("handle_cmnd: read returned %d\n", err);
								break;
							}
//                          if (err > 0)
//                              inode_dir_notify (dev_file->f_dentry->d_parent->d_inode, DN_MODIFY);
						}
					}

					/*
					 * Assumption is that everything will be
					 * written in blocksize chunks. Also the
					 * blocksize is less than pagesize and
					 * the latter is a multiple of the first
					 */
/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
					lba_start += sg_dma_len(st_list + i) / BLOCKSIZE;
#else
					lba_start += st_list[i].length / BLOCKSIZE;
#endif
				}
				cmnd->state = ST_DONE;
				cmnd->req->sr_result = DID_OK << 16;
			}
			err = 0;
			break;
		}

	default:
		{
# ifdef DEBUG_HANDLE_CMD
			printk("unknown command received\n");
# endif
			break;
		}
	}
	return err;
}

#endif

/* dispatch_scsi_info is the central dispatcher 
 * It is the interface between the proc-fs and the SCSI subsystem code
 */
static int
proc_scsi_target_read(char *buffer, char **start, off_t offset, int length,
					  int *eof, void *data)
{
	Scsi_Target_Device *tpnt = data;
	int n = 0;

	/*****
	printk("Entering proc_scsi_target_read\n");
	printk("offset: %d, length %d, tpntid: %d\n", (int) offset, length, 0);
	*****/

	if (tpnt->template->proc_info == NULL) {
/*		n = generic_proc_info(buffer, start, offset, length,
				      hpnt->hostt->info, hpnt); */
	} else {
		n = (tpnt->template->proc_info(buffer, start, offset, length, 0, 0));
	}
	*eof = (n < length);
	return n;
}

#define PROC_BLOCK_SIZE (3*1024)	/* 4K page size, but our output routines 
									   * use some slack for overruns 
									 */

static int
proc_scsi_target_write(struct file *file, const char *buf,
					   unsigned long count, void *data)
{
	Scsi_Target_Device *tpnt = data;
	ssize_t ret = 0;
	char *page;
	char *start;

	if (count > PROC_BLOCK_SIZE)
		return -EOVERFLOW;

	if (!(page = (char *) __get_free_page(GFP_KERNEL)))
		return -ENOMEM;
	copy_from_user(page, buf, count);

	if (tpnt->template->proc_info == NULL)
		ret = -ENOSYS;
	else
		ret = tpnt->template->proc_info(page, &start, 0, count, tpnt->id, 1);
	free_page((ulong) page);
	return ret;
}

/* Define the function that handles the proc interface - Added for *
 * configuring the login parameters on the iSCSI target side       */

#ifdef CONFIG_PROC_FS

static int
scsi_target_proc_info(char *buffer, char **start, off_t offset, int length)
{

	return sprintf(buffer, "Function to be implemented\n");
}

static int
proc_scsi_target_gen_write(struct file *file, const char *buf,
						   unsigned long length, void *data)
{

	return 0;
}

static void
build_proc_target_dir_entries(Scsi_Target_Device * the_device)
{
	struct proc_dir_entry *p;
	char name[10];

	/* create the proc directory entry for the device */
	the_device->template->proc_dir =
		proc_mkdir(the_device->template->name, proc_scsi_target);

	/* create the proc file entry for the device */
	sprintf(name, "%d", (int) the_device->id);
	p = create_proc_read_entry(name, S_IFREG | S_IRUGO | S_IWUSR,
							   the_device->template->proc_dir,
							   proc_scsi_target_read, (void *) the_device);
	if (!p)
		panic
			("Not enough memory to register SCSI target in /proc/scsi_target\n");
	p->write_proc = proc_scsi_target_write;

}

#endif

#ifdef MODULE
MODULE_AUTHOR("Bob Russell");
MODULE_DESCRIPTION("UNH");
MODULE_SUPPORTED_DEVICE("sd");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#endif

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
EXPORT_SYMBOL(target_num);
EXPORT_SYMBOL(scsi_release);
EXPORT_SYMBOL(scsi_target_done);
EXPORT_SYMBOL(scsi_rx_data);
EXPORT_SYMBOL(register_target_template);
EXPORT_SYMBOL(deregister_target_template);
EXPORT_SYMBOL(register_target_front_end);
EXPORT_SYMBOL(deregister_target_front_end);
EXPORT_SYMBOL(rx_task_mgmt_fn);
EXPORT_SYMBOL(rx_cmnd);
#endif
