/*	target/target_error_rec.c 

    vi: set autoindent tabstop=4 shiftwidth=4

*/

/*
	Copyright (C) 2001-2003 InterOperability Lab (IOL)
	University of New Hampshier (UNH)
	Durham, NH 03824

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
	USA.

	The name of IOL and/or UNH may not be used to endorse or promote products
	derived from this software without specific prior written permission.
*/

#include "iscsi_target.h"
#include "target_error_rec.h"

static struct semaphore targ_sess_recovery_sem;

/************************************************************************
 *  The Main Error Recovery Handler Called from iscsi_initiator when it *
 *  finds any error.                                                    *
 *  Returns 0 if recovery completed ok, -1 otherwise                    *
 ************************************************************************/

int
targ_do_error_recovery(struct targ_error_rec *err_rec)
{

	int retval = 0;
	struct iscsi_conn *curr_conn;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_do_error_recovery\n");

	if (err_rec == NULL) {
		TRACE_ERROR("Error Recovery structure is NULL\n");
		retval = -1;
		goto out;
	}

	if (err_rec->curr_conn == NULL) {
		TRACE_ERROR("Connection Pointer is NULL in Error Recovery structure\n");
		retval = -1;
		goto out;
	}

	curr_conn = err_rec->curr_conn;

	switch (curr_conn->session->oper_param->ErrorRecoveryLevel) {
	case SESSION_RECOVERY:
		init_MUTEX_LOCKED(&targ_sess_recovery_sem);
		targ_session_recovery(curr_conn);
		break;
	case DIGEST_RECOVERY:
		retval = targ_digest_recovery(err_rec);
		break;
	case CONNECTION_RECOVERY:
		TRACE_ERROR("FUNCTION Not Implemented yet\n");
		retval = -1;
		break;
	}

out:
	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_do_error_recovery, retval %d\n",
		retval);
	return retval;
}

/************************************************************************
* This function handles the session recovery. It creates a recovery     *
* thread that will do session recovery.                                 *
************************************************************************/

void
targ_session_recovery(struct iscsi_conn *current_connection)
{
	TRACE(TRACE_ENTER_LEAVE, "Enter targ_session_recovery\n");

	TRACE(TRACE_ISCSI,
		  "Executing Target Session Recovery - cancelling Receive Thread\n");

	/* Create a recovery thread */
	kernel_thread((int (*)(void *)) targ_recovery_thread,
				  (void *) current_connection, 0);

	/* Wait till the recovery thread completes processing */
	down_interruptible(&targ_sess_recovery_sem);

	TRACE(TRACE_ENTER_LEAVE, "Leave targ_session_recovery\n");

	return;
}

/************************************************************************
* This thread handles the session recovery. Session Recovery will       *
* abort all pending tasks and closes all transport connection           *
* Section (6.12.4)                                                      *
************************************************************************/

void
targ_recovery_thread(void *arg)
{
	struct iscsi_conn *current_connection;
	struct iscsi_session *current_session;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_recovery_thread\n");

	current_connection = (struct iscsi_conn *) arg;
	current_session = current_connection->session;

	TRACE(TRACE_ISCSI, "target_recovery_thread started\n");

	if (!down_interruptible(&current_session->devdata->session_sem)) {
		iscsi_release_session(current_session);

		up(&targ_sess_recovery_sem);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_recovery_thread\n");

	return;
}

/************************************************************************
 * This function handles the digest errors and sequence errors that may *
 * have caused because of previous digest errors.                       *
 * Returns 0 if recovery is completed correctly, -1 otherwise           *
 ************************************************************************/

int
targ_digest_recovery(struct targ_error_rec *err_rec)
{
	int retval = 0, data_length = 0;
	int size = 0, err_type = 0, data_offset = 0;
	unsigned int opcode = 0;
	struct iscsi_conn *curr_conn;
	struct iscsi_init_scsi_data_out *data_out_hdr;
	struct iscsi_cmnd *cmd;

	curr_conn = err_rec->curr_conn;
	err_type = err_rec->err_type;
	cmd = err_rec->cmd;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_digest_recovery\n");

	if (err_rec->pdu_hdr == NULL) {
		TRACE_ERROR("iscsi PDU  Header is NULL\n");
		retval = -1;
		goto digest_out;
	} else {
		opcode = err_rec->pdu_hdr->opcode;
		size = err_rec->pdu_hdr->length;
		if (size > 0xffffff) {
			TRACE_ERROR("TotalAHSLength 0x%02x, expected 0x00\n", size >> 24);
			retval = -1;
			goto digest_out;
		}
	}

	if (err_type == HEADER_DIGERR) {
		if (size > 0) {
			retval = targ_drop_pdu_data(curr_conn, size);
		}
		goto digest_out;
	}

	if (opcode == ISCSI_INIT_SCSI_DATA_OUT) {
		if (cmd == NULL) {
			/* TRACE_ERROR("Related SCSI Command is NULL\n); */
			retval = -1;
			goto digest_out;
		}

		if (err_type == PAYLOAD_DIGERR) {
			iscsi_tx_rjt(curr_conn, REASON_DATA_DIGEST_ERR);
			send_recovery_r2t(cmd, cmd->data_done, NULL);
		} else if (err_type == SEQUENCE_ERR) {
			/* error recovery ver ref18_04 - SAI */
			/* if there is a sequence error and the arrived data 
			 * seq. indicates lost packets, queue the received 
			 * packets and request re-transmission of lost packets 
			 */
			queue_data(err_rec);

			data_length = cmd->data_length - cmd->data_done;
			data_offset = cmd->data_length - data_length;

			data_out_hdr = (struct iscsi_init_scsi_data_out *) err_rec->pdu_hdr;
			size = (data_out_hdr->offset - data_offset);

			send_recovery_r2t(cmd, data_offset, NULL);

			//up(&curr_conn->tx_sem);

		} else {
			TRACE_ERROR("Unknown err_type %d for DataOut "
					"recovery\n", err_type);
			retval = -1;
		}

	} else if (opcode == ISCSI_INIT_SCSI_CMND) {
		if (err_type == PAYLOAD_DIGERR) {
			/* digest error in immediate data attached to write */
			iscsi_tx_rjt(curr_conn, REASON_DATA_DIGEST_ERR);
		} else {
			TRACE_ERROR("Unknown err_type %d for SCSI_CMND "
					"recovery\n", err_type);
			retval = -1;
		}
	} else {
		TRACE_ERROR("Unknown err_type %d for recovery\n", err_type);
		retval = -1;
	}

digest_out:

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_digest_recovery, retval %d\n",
		retval);

	return retval;
}

/*******************************************************************
 * Reads the data portion of the dropped PDU due to digest errors. *
 * The data read is also ignored.                                  *
********************************************************************/

int
targ_drop_pdu_data(struct iscsi_conn *curr_conn, int size)
{
	int padding = 0, retval = 0, data_length = 0;
	int total_rx = 0, rx_loop = 0;
	char *buffer = NULL;
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;

	TRACE(TRACE_ENTER_LEAVE, "Entering targ_drop_pdu_data\n");

	padding = -(size) & 3;
	size += padding;

	if (curr_conn->data_crc == 1) {
		size += CRC_LEN;
	}

	TRACE(TRACE_DEBUG, "targ_drop_pdu: data_read %d total_data %d\n",
		  data_length, size);

	data_length = size;

	while (data_length > 0) {
		if (data_length > MAX_MALLOC_SIZE)
			data_length = MAX_MALLOC_SIZE;

		buffer = kmalloc(data_length * sizeof (char), GFP_KERNEL);
		if (!buffer) {
			TRACE_ERROR("Malloc Error : Unable to drop the data PDU\n");
			retval = -ENOMEM;
			goto end_drop_pdu;
		}

		for (total_rx = 0; total_rx < data_length;) {
			memset(&msg, 0, sizeof (struct msghdr));
			msg.msg_iovlen = 1;
			msg.msg_iov = &iov;

			iov.iov_len = data_length - total_rx;
			iov.iov_base = buffer + total_rx;

			oldfs = get_fs();
			set_fs(get_ds());
			rx_loop = sock_recvmsg(curr_conn->conn_socket, &msg,
								   (data_length - total_rx), MSG_WAITALL);
			set_fs(oldfs);

			if (rx_loop <= 0) {
				TRACE_ERROR("trag_drop_pdu: rx_loop %d total_rx %d\n", rx_loop,
							total_rx);
				retval = -ECONNRESET;	/* Connection reset by peer */
				kfree(buffer);
				goto end_drop_pdu;
			}
			total_rx += rx_loop;

			TRACE(TRACE_DEBUG, "targ_drop_pdu: rx_loop %d total_rx %d\n",
				  rx_loop, total_rx);
		}

		size -= data_length;
		data_length = size;

		TRACE(TRACE_DEBUG,"targ_drop_pdu: data_read %d total_data %d\n",
			  data_length, size);

		kfree(buffer);
	}

end_drop_pdu:

	TRACE(TRACE_ENTER_LEAVE, "Leaving targ_drop_pdu_data\n");

	return retval;
}

void deal_with_r2t_timer(unsigned long data);

/*********************************************************************
 * (re)start R2T Timer is used for tracking the R2T sent. This timer
 * helps re-transmit the most recent R2T sent to the initiator.
 ********************************************************************/
/* (re)start the R2T retransmit timer */
void
restart_r2t_timer(struct iscsi_session *session)
{
	__u32 period = session->r2t_period;

	TRACE(TRACE_ENTER_LEAVE, "Enter restart_r2t_timer period %u\n", period);

	if (period && session->r2t_timer && !timer_pending(session->r2t_timer)){
		TRACE(TRACE_DEBUG, "Start r2t timer for %u ticks\n", period);
		session->r2t_timer->expires = jiffies + period;
		session->r2t_timer->data = (unsigned long) session;
		session->r2t_timer->function = deal_with_r2t_timer;
		add_timer(session->r2t_timer);
	} else {
		TRACE_ERROR("R2T timer not started for %u ticks\n", period);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave restart_r2t_timer\n");
}

/*********************************************************************
* Called by kernel when R2T periodic session timer expires.
*********************************************************************/
void
deal_with_r2t_timer(unsigned long data)
{
	struct iscsi_session *session;
	struct iscsi_cmnd *cmnd = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter deal_with_r2t_timer\n");

	TRACE(TRACE_DEBUG, "R2T retransmit timer expired\n");

	session = (struct iscsi_session *) data;

	for (cmnd = session->cmnd_list; cmnd != NULL; cmnd = cmnd->next) {
		if (cmnd->outstanding_r2t > 0 &&
			(jiffies > (cmnd->timestamp + session->r2t_period))) {
			up(&session->retran_sem);
			break;
		}
	}

	restart_r2t_timer(session);

	TRACE(TRACE_ENTER_LEAVE, "Leave deal_with_r2t_timer\n");
}

/***************************************************************************
* create an R2T cookie and store the R2T details in the cookie. This cookie
* is esssential for R2T re-transmssions in error conditions.
***************************************************************************/
struct iscsi_cookie *
create_r2t_cookie(struct iscsi_cmnd *cmnd)
{
	struct iscsi_cookie *cookie = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter create_r2t_cookie\n");

	cookie =
		(struct iscsi_cookie *) kmalloc(sizeof (struct iscsi_cookie),
										GFP_KERNEL);
	if (cookie == NULL) {
		TRACE_ERROR("Malloc problem in creating R2T cookie\n");
		goto end_iscsi_cookie;
	}
	cookie->next = NULL;

	if (cmnd->first_r2t_cookie == NULL) {
		cmnd->first_r2t_cookie = cmnd->last_r2t_cookie = cookie;
	} else {
		cmnd->last_r2t_cookie->next = cookie;
		cmnd->last_r2t_cookie = cookie;
	}

  end_iscsi_cookie:

	TRACE(TRACE_ENTER_LEAVE, "Leave create_r2t_cookie\n");

	return cookie;
}

/* free the long pending R2T cookie list after the command is completed - SAI */
void
free_r2t_cookie(struct iscsi_cmnd *cmnd)
{
	struct iscsi_cookie *cookie = cmnd->first_r2t_cookie;
	struct iscsi_cookie *tmp_cookie = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter free_r2t_cookie\n");

	while (cookie) {
		tmp_cookie = cookie;
		cookie = cookie->next;
		kfree(tmp_cookie);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave free_r2t_cookie\n");

	return;

}

/**************************************************************************
* Re-Transmits the R2T indicated by the r2tsn. This function is invoked by
* SNACK requests and also by local R2T timeouts.
**************************************************************************/
int
retran_r2t(struct iscsi_cmnd *cmnd, int r2tsn)
{
	int retval = -1;  /* pessimistic assumption that the R2T will 
							not be found */
	struct iscsi_cookie *cookie = cmnd->first_r2t_cookie;

	TRACE(TRACE_ENTER_LEAVE, "Enter retran_r2t\n");

	while (cookie) {
		if (cookie->seq == r2tsn) {
			send_recovery_r2t(cmnd, 0, cookie);
			retval = 1;
			break;
		}
		cookie = cookie->next;
	}

	if (retval == -1) {
		TRACE_ERROR
			("Unidentified R2T seq. no %d retransmission request received \n",
			 r2tsn);
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave retran_r2t\n");

	return retval;
}

/*******************************************************************************
* This function uses the offset and data length parameter or cookie information
* to build a recovery R2T to be sent to the intiator for missing Dataouts. This
* function is called by SNACK mechanism and also by other places of the code that
* finds a discrepency in Dataouts using data sequence numbering or offsets.
********************************************************************************/
int
send_recovery_r2t(struct iscsi_cmnd *cmnd, int data_offset,
				  struct iscsi_cookie *cookie)
{
	int data_length = 0;
	int max_burst_len = 0;

	TRACE(TRACE_ENTER_LEAVE, "Enter send_recovery_r2t\n");

	if (cmnd->recovery_r2t)
		/* don't send recovery_r2t for another error */
		goto out;
	else
		cmnd->recovery_r2t = 1;

	cmnd->state = ISCSI_BUFFER_RDY;
	max_burst_len = cmnd->conn->session->oper_param->MaxBurstLength;
	data_length = cmnd->data_length - data_offset;

	if (!cookie) {
		cmnd->r2t_data = data_length;
		cookie = create_r2t_cookie(cmnd);
		if (cookie) {
			cookie->seq = cmnd->r2t_sn;
			cookie->offset = data_offset;
			cookie->xfer_len =
				(data_length <= max_burst_len) ? data_length : max_burst_len;
		}
	} else {
		cmnd->r2t_data = cmnd->data_length - cookie->offset;
		cmnd->r2t_sn = cookie->seq;
		cmnd->outstanding_r2t--;
		up(&cmnd->conn->tx_sem);
	}

	TRACE(TRACE_ISCSI, "Send recovery R2T for ITT %u\n",
		cmnd->init_task_tag);

  out:
	TRACE(TRACE_ENTER_LEAVE, "Leave send_recovery_r2t\n");
	return (0);
}

/***********************************************************************
* This thread remains blocked at a semaphore for most part of it's time
* When a r2t timeout happens this thread is awakened for re-transmitting
* the r2t request.
***********************************************************************/

void
iscsi_retran_thread(void *param)
{
	struct iscsi_session *session = (struct iscsi_session *) param;
	struct iscsi_cmnd *cmnd = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Enter iscsi_retran_thread\n");

	lock_kernel();

/* Ming Zhang, mingz@ele.uri.edu */
#ifdef K26
	daemonize(
#else
	daemonize();
	sprintf(current->comm,
#endif

		"iscsi_tran_%d", session->devdata->ntsih);

	siginitsetinv(&current->blocked, ISCSI_SHUTDOWN_SIGBITS);

	session->retran_thread = current;

	unlock_kernel();

	while (1) {
		if (down_interruptible(&session->retran_sem))
			break;

		for (cmnd = session->cmnd_list; cmnd; cmnd = cmnd->next) {
			if (cmnd->outstanding_r2t > 0 &&
				(jiffies >
				 (cmnd->timestamp + session->r2t_period))) {
				retran_r2t(cmnd, cmnd->r2t_sn - 1);
			}
		}
	}

	TRACE(TRACE_ENTER_LEAVE, "Leave iscsi_retran_thread\n");

	session->retran_thread = NULL;
	up(&session->thr_kill_sem);
}

void
add_data_to_queue(struct iscsi_cmnd *cmd, struct iscsi_cookie *dataq)
{
	TRACE(TRACE_ENTER_LEAVE, "Entering add_data_to_queue\n");

	if (!cmd->first_data_q) {
		cmd->first_data_q = dataq;
		cmd->last_data_q = dataq;
	} else {
		cmd->last_data_q->next = dataq;
		cmd->last_data_q = cmd->last_data_q->next;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving add_data_to_queue\n");

	return;
}

/********************************************************************
* This function drives the out-of-order dataouts received into the 
* appropriate scatter gather buffer location maintained by scsi. The
* information about the out-of-order dataout and also the corresponding
* scatter offsets are queued for future use.
**********************************************************************/
int
queue_data(struct targ_error_rec *err_rec)
{
	int offset = 0, size = 0, list_offset = 0;
	int count = 0;
	int niov, padding, err, retval = 0, data_crc_len = 0;
	__u32 data_sn, digest, pad_bytes;

	struct iscsi_conn *conn = NULL;
	struct iscsi_cmnd *cmd;
	struct iscsi_init_scsi_data_out *hdr;
	struct iscsi_cookie *dataq = NULL;
	struct scatterlist *st_list;
	struct iovec *iov = NULL;

	TRACE(TRACE_ENTER_LEAVE, "Entering queue_data\n");

	cmd = err_rec->cmd;
	hdr = (struct iscsi_init_scsi_data_out *) err_rec->pdu_hdr;

	list_offset = offset = hdr->offset;
	data_sn = hdr->data_sn;
	size = hdr->length;
	conn = err_rec->curr_conn;

	st_list = (struct scatterlist *) cmd->cmnd->req->sr_buffer;

	while (st_list[count].length <= list_offset) {
		list_offset -= st_list[count].length;
		count++;
	}

	st_list += count;
	niov = find_iovec_needed(size, st_list, list_offset);

	if (niov < 0) {
		TRACE_ERROR("queue data: find_iovec_needed returned an error\n");
		return -1;
	}

	padding = -(size) & 3;
	if (padding)
		niov++;

	if (conn->data_crc == 1)
		niov++;

	iov = (struct iovec *) kmalloc(niov * sizeof (struct iovec), GFP_KERNEL);

	if (!iov) {
		TRACE_ERROR("queue_data: Could not malloc iov\n");
		return -1;
	}

	TRACE(TRACE_DEBUG, "queue_data: no. of iovec needed %d\n", niov);

	retval = fill_iovec(iov, 0, niov, st_list, &list_offset, size);

	if (padding) {
		iov[niov - 1 - conn->data_crc].iov_base = &pad_bytes;
		iov[niov - 1 - conn->data_crc].iov_len = padding;
	}

	if (conn->data_crc == 1) {
		data_crc_len = CRC_LEN;
		iov[niov - 1].iov_base = &digest;
		iov[niov - 1].iov_len = CRC_LEN;
	}

	err = iscsi_rx_data(cmd->conn, iov, niov, size + padding + data_crc_len);

	if (err != (size + padding + data_crc_len)) {
		TRACE_ERROR("queue_data: Trouble in iscsi_rx_data\n");
		if (err == PAYLOAD_DIGERR)
			return err;

	}

	dataq = kmalloc(sizeof (struct iscsi_cookie), GFP_KERNEL);

	if (!dataq) {
		TRACE_ERROR("add_data_to_queue: Malloc returned an error\n");
		return -1;
	}

	dataq->offset = offset;
	dataq->xfer_len = size;
	dataq->list_offset = list_offset;
	dataq->list_count = retval;
	dataq->next = NULL;

	add_data_to_queue(cmd, dataq);

	TRACE(TRACE_ENTER_LEAVE, "Leaving queue_data\n");

	return 0;
}

/***********************************************************************
* This function is called everytime a in-sequence dataout pdu is
* received. The function verifies if there was any previous
* out-of-order dataouts for the command and updates the state variables
* accordingly.
***********************************************************************/
void
search_data_q(struct iscsi_cmnd *cmd)
{
	int unhook_flg = 0;
	struct iscsi_cookie *dataq, *prev, *tmp_q;

	TRACE(TRACE_ENTER_LEAVE, "Entering search_data_q\n");

	if (!cmd->first_data_q)
		return;

	dataq = prev = cmd->first_data_q;

	while (dataq) {
		if (cmd->data_done > dataq->offset)
			unhook_flg = 1;
		else if (cmd->data_done == dataq->offset) {
			cmd->scatter_list_count += dataq->list_count;
			cmd->scatter_list_offset = dataq->list_offset;
			cmd->data_done += dataq->xfer_len;
			unhook_flg = 1;
		}

		if (unhook_flg) {
			/* unhook data node */
			if (prev == cmd->first_data_q) {
				cmd->first_data_q = dataq->next;
				prev = dataq->next;
			} else if (dataq == cmd->last_data_q)
				cmd->last_data_q = prev;
			else
				prev->next = dataq->next;

			tmp_q = dataq;
			dataq = dataq->next;
			kfree(tmp_q);
			unhook_flg = 0;
		} else {
			prev = dataq;
			dataq = dataq->next;
		}

		if (cmd->data_done >= cmd->data_length)
			break;
	}

	TRACE(TRACE_ENTER_LEAVE, "Leaving search_data_q\n");
}
